"""
Module for the plpipes package.

This module serves as a container for the plpipes library. All the
interesting functionalities and features are implemented in the 
submodules of this package.
"""

from plpipes.config import cfg
import logging



import logging
import pandas
import sqlalchemy.sql as sas

from plpipes.database.backend import Backend
from plpipes.database.sqlext import AsSubquery, Wrap
from plpipes.util.database import split_table_name

DEFAULT_CHUNKSIZE = 5000

class PandasBackend(Backend):
    def query(self, txn, sql, parameters, kws):
        return self._df_read_sql(txn, Wrap(sql), params=parameters, **kws)

    def query_chunked(self, txn, sql, parameters, kws):
        chunksize = txn._driver._pop_kw(kws, "chunksize", DEFAULT_CHUNKSIZE)
        for chunk in self._df_read_sql(txn, Wrap(sql), params=parameters, chunksize=chunksize, **kws):
            yield chunk

    def query_group(self, txn, sql, parameters, by, kws):
        if by is None or not by:
            raise ValueError("by argument must contain a list of column names")
        wrapped_sql = sas.select("*").select_from(AsSubquery(Wrap(sql))).order_by(*[sas.column(c) for c in by])

        tail = None
        for chunk in self.query_chunked(txn, wrapped_sql, parameters, kws):
            if tail is not None:
                chunk = self._df_concat([tail, chunk])
            groups = [g for _, g in chunk.groupby(by)]
            tail = groups.pop()
            for group in groups:
                group = group.reset_index()
                yield group
        if tail is not None:
            yield tail.reset_index()

    def register_handlers(self, handlers):
        handlers["create_table"].register(pandas.DataFrame, self._create_table_from_pandas)

    def _create_table_from_pandas(self, txn, table_name, df, paramaters, if_exists, kws):
        logging.debug(f"Creating table {table_name} from pandas dataframe (shape: {df.shape})")
        chunksize = txn._driver._pop_kw(kws, "chunksize", DEFAULT_CHUNKSIZE)
        schema, table_name = split_table_name(table_name)
        df.to_sql(table_name, txn._conn,
                  schema=schema, if_exists=if_exists,
                  index=False, chunksize=chunksize,
                  **kws)

    def create_table_from_records(self, txn, table_name, records, paramaters, if_exists, kws):
        df = pandas.DataFrame.from_records(records)
        self._create_table_from_pandas(txn, table_name, df, paramaters, if_exists, kws)

    def _df_read_sql(self, txn, sqla, **kws):
        return pandas.read_sql(sqla, txn._conn, **kws)

    def _df_concat(self, dfs):
        return pandas.concat(dfs)



import logging
import sqlalchemy.sql as sas

from plpipes.database.backend import Backend
from plpipes.database.sqlext import AsSubquery, Wrap

class DictBackend(Backend):
    def query(self, txn, sql, parameters, kws):
        result = txn._conn.execute(Wrap(sql), parameters=parameters, **kws)
        return [{**row._mapping} for row in result.all()]

    def query_first(self, txn, sql, parameters, kws):
        for row in txn._conn.execute(Wrap(sql), parameters=parameters, **kws):
            return {**row._mapping}

    def register_handlers(self, handlers):
        # TODO
        # handlers["create_table"].register(list), self._create_table_from_list)
        # handlers["create_table"].register(dict), self._create_table_from_dict)
        pass



import pandas as pd
import httpx
import datetime
import logging

from plpipes.config import cfg

_url_v2="https://api.bls.gov/publicAPI/v2/timeseries/data/"

def _get_series(series, start_year=1957, end_year=None):
    if end_year is None:
        end_year = datetime.datetime.now().year
    dfs = []
    year = start_year
    api_key = cfg["net.client.us_bls.api_key"]
    while year <= end_year:
        top_year = min(year + 19, end_year)
        logging.debug(f"Retrieving series {series} data from {year} to {top_year} from US BLS")
        r = httpx.post(_url_v2, json={"seriesid": [series],
                                      "startyear": str(year),
                                      "endyear": str(top_year),
                                      "registrationkey": api_key })
        df = pd.DataFrame(r.json()["Results"]["series"][0]["data"])
        dfs.append(df)
        year += 20

    df = pd.concat(dfs[::-1], axis=0, ignore_index=True)
    df.columns = df.columns.str.replace('-', '_')
    return df.reset_index(drop=True)

def us_inflation(**kwargs):
    df = _get_series("CUUR0000SA0L1E", **kwargs)
    df["month"] = df['period'].str[1:].astype(int)
    df['date'] = pd.to_datetime(df['year'].astype(str) + '-' + df['month'].astype(str) + '-01').dt.date
    df = df.sort_values(by='date')
    return df.reset_index(drop=True)





from plpipes.plugin import plugin
from plpipes.cloud.azure.auth.base import AuthenticatorBase
from azure.identity import ClientSecretCredential
import logging

from plpipes.exceptions import AuthenticationError


@plugin
class CSAuthenticator(AuthenticatorBase):

    def _authenticate(self):
        return ClientSecretCredential(tenant_id=self._cfg["tenant_id"],
                                      client_id=self._cfg["client_id"],
                                      client_secret=self._cfg["client_secret"])



import logging
import sqlalchemy.engine
from urllib.parse import urlunparse, urlparse

from plpipes.database.driver.sqlalchemy import SQLAlchemyDriver
from plpipes.plugin import plugin

def first(*args):
    for a in args:
        if a:
            return a
    return None

_default_sqla_subdriver = { "mysql": "pymysql",
                            "mariadb": "mariadbconnector" }

def first(*args):
    for a in args:
        if a:
            return a
    return None

class MySQLBaseDriver(SQLAlchemyDriver):

    def _default_sqla_driver(self):
        return "mysql"

    def __init__(self, name, drv_cfg):
        dsad = self._default_sqla_driver()
        cs = urlparse(drv_cfg.get("connection_string", f"{dsad}:"))
        if cs.scheme is None or cs.scheme == '':
            sqla_driver = dsad
            more = []
        else:
            sqla_driver, *more = cs.scheme.split("+", 2)

        sqla_subdriver = first(*more, _default_sqla_subdriver.get(sqla_driver))
        if sqla_subdriver is not None:
            sqla_driver += "+" + sqla_subdriver

        host     = drv_cfg.setdefault('host',     first(cs.hostname))
        port     = drv_cfg.setdefault('port',     first(cs.port))
        database = drv_cfg.setdefault('database', first(cs.path))
        user     = drv_cfg.setdefault('user',     first(cs.username))
        password = drv_cfg.setdefault('password', first(cs.password))

        sqla_driver = drv_cfg.setdefault('sqlalchemy_driver', sqla_driver)

        url = sqlalchemy.engine.URL.create(sqla_driver,
                                           username=user, password=password,
                                           host=host, port=port,
                                           database=database)

        logging.debug(f"SQLAlchemy MySQL (or derivative) url: {url}")

        super().__init__(name, drv_cfg, url)



import pandas as pd
import httpx

from plpipes.config import cfg

def petroleum_prices():
    r = httpx.get("https://api.eia.gov/v2/petroleum/pri/spt/data/?frequency=weekly&data[0]=value&facets[product][]=EPCBRENT&facets[product][]=EPCWTI&sort[0][column]=period&sort[0][direction]=desc",
                  timeout=120,
                  params = {'api_key': cfg['net.client.eia.api_key']})
    df = pd.DataFrame(r.json()['response']['data'])
    df.columns = df.columns.str.replace('-', '_')
    df['period'] = pd.to_datetime(df['period']).dt.date
    return df




from plpipes.config import cfg

PLURALS = cfg.to_tree("util.pluralsingular.plurals")
SINGULARS = {lang: {v: k for k, v in plurals.items()} for lang, plurals in PLURALS.items()}

LANG = cfg.get("util.pluralsingular.lang", "en")

def pluralize(word, lang=LANG, marks=True, plurals=None):
    if plurals is None:
        plurals = PLURALS.get(lang, {})
    else:
        plurals = {**PLURALS.get(lang, {}), **plurals}
    if isinstance(words, list):
        return [_pluralize_word(w, lang, marks, plurals) for w in words]
    else:
        return _pluralize_word(words, lang, marks, plurals)

def _pluralize_word(word, lang, marks, plurals):
    if word in plurals:
        p = plurals[word]
    else:
        import pluralsingular
        p = pluralsingular.pluralize(word, lang=lang)
    if not marks:
        import unidecode
        p = unidecode.unidecode(p)
    return p

def singularize(words, lang=LANG, marks=True, singulars=None):
    if singulars is None:
        singulars = SINGULARS.get(lang, {})
    else:
        singulars = {**SINGULARS.get(lang, {}), **singulars}
    if isinstance(words, list):
        return [_singularize_word(w, lang, marks, singulars) for w in words]
    else:
        return _singularize_word(words, lang, marks, singulars)

def _singularize_word(word, lang, marks, singulars):
    if word in singulars:
        s = singulars[word]
    else:
        import pluralsingular
        s = pluralsingular.singularize(word, lang=lang)
    if not marks:
        import unidecode
        s = unidecode.unidecode(s)
    return s



from plpipes.plugin import plugin
import plpipes.database as db
import pathlib

from plpipes.tool.dbeaver.conarg import ConArg

@plugin
class SQLiteConArg(ConArg):
    def __init__(self, name, db_drv):

        super().__init__(name, db_drv)
        self._fn = fn = db_drv.backing_filename()
        self.database = str(pathlib.Path(fn).absolute())
        self.driver = "sqlite"

    def active(self):
        if pathlib.Path(self._fn).exists():
            return True

    def __str__(self):
        args = []
        for arg in self._con_args:
            if hasattr(self, arg):
                args.append(f"{arg}={getattr(self, arg)}")
        return "|".join(args)



from plpipes.plugin import plugin, Plugin

from openai import OpenAI

@plugin
class OpenAIProvider(Plugin):

    @staticmethod
    def get_client(name, cfg):
        api_key = cfg["api_key"]
        return OpenAI(api_key = api_key)




from plpipes.database.driver.azure_sql import AzureSQLDriver
from plpipes.plugin import plugin

plugin(AzureSQLDriver)

# TODO: Having the code in parallel clases is not really required. Move it here!



"""
This module implements the basic drivers for running Python code and sequences of actions
in the plpipes framework.

This module contains action drivers for executing Python scripts and for sequencing
multiple actions. The action drivers are responsible for managing the execution of
specific actions defined in the project.

"""

import logging

from plpipes.config import cfg
from plpipes.action.base import Action
from plpipes.action.registry import register_class
from plpipes.action.runner import lookup

import plpipes.database
import plpipes

def _action_namespace_setup(action_cfg=None):
    """
    Set up the action namespace for executing Python code.

    Args:
        action_cfg: Optional configuration specific to the action.

    Returns:
        A dictionary containing the configuration, action configuration, database, and plpipes module.
    """
    return {"cfg": cfg, "action_cfg": action_cfg, "db": plpipes.database, "plpipes": plpipes}

class _PythonRunner(Action):
    """
    Class to run Python scripts as actions in the plpipes framework.

    Inherits from the Action base class and overrides the method to execute the script
    code defined in the action configuration.
    """

    def _do_it(self, indent):
        """
        Execute the Python script defined in the action configuration.

        Args:
            indent: An indentation level for logging purposes.

        Raises:
            Exception: Raises an exception if there is an error during compilation or execution of the script.
        """
        if not hasattr(self, "_code"):
            self._path = self._cfg["files.py"]
            try:
                with open(self._path, "r", encoding="utf8") as f:
                    py_code = f.read()
                self._code = compile(py_code, self._path, 'exec')
            except Exception as ex:
                logging.error(f"Action of type python_script failed while compiling {self._path}")
                raise ex
        try:
            logging.debug(f"Running python code at {self._path}")
            exec(self._code, _action_namespace_setup(action_cfg=self._cfg))
            del self._code
        except Exception as ex:
            logging.error(f"Action of type python_script failed while executing {self._path}")
            raise ex

class _Sequencer(Action):
    """
    Class to execute a sequence of actions defined in the action configuration.

    Inherits from the Action base class and provides functionality to run
    multiple child actions in the order specified in the configuration.
    """

    def do_it(self):
        """
        Execute the sequence of actions as defined in the action configuration.

        This method retrieves the names of the child actions and runs them in order.

        Raises:
            Exception: Raises an exception if any of the child actions fail to execute.
        """
        name = self._name

        for child_name in self._cfg["sequence"]:
            lookup(child_name, parent=name).run()

register_class("python_script", _PythonRunner, "py")
register_class("sequence", _Sequencer, "dir")



import logging
import pathlib
import importlib

from plpipes.action.base import Action
from plpipes.action.registry import register_class

class _Downloader(Action):
    def do_it(self):
        print(f"Action config: {self._cfg.to_tree()}")
        service_name = self._cfg["service"]
        path = self._cfg.get("path", "")
        params = self._cfg.to_tree("params")
        print(f"params: {params}")
        db = self._cfg.get("db", "work")

        impl = importlib.import_module(f"plpipes.action.driver.downloader.service.{service_name.lower()}")
        service = impl.download(path=path, params=params, db=db, acfg=self._cfg)

register_class("downloader", _Downloader)




def optional_abstract(method):
    def wrapper(*args, **kwargs):
        raise NotImplementedError(f"Optional method {method.__name__} is not implemented by the class {args[0].__class__.__name__}")
    wrapper.not_implemented = True
    return wrapper



import pandas as pd
import io
import requests

def eur_usd():
    url = "https://data-api.ecb.europa.eu/service/data/EXR/D.USD.EUR.SP00.A?format=csvdata"
    s = requests.get(url).content
    df = pd.read_csv(io.StringIO(s.decode('utf-8')),
                     parse_dates=["TIME_PERIOD"])

    return df

    




import sys
from plpipes.config import cfg
import plpipes.filesystem as fs
import plpipes.database as db
import plpipes.plugin
import logging
import findapp
import subprocess
import os

_conarg_class_registry = plpipes.plugin.Registry("dbeaver_conarg_backend", "plpipes.tool.dbeaver.conarg.driver")

def _conarg_lookup(name):
    db_drv = db.lookup(name)
    try:
        plugin_name = db_drv.driver_name()
        conarg_class = _conarg_class_registry.lookup(plugin_name)
        return conarg_class(name, db_drv)
    except ModuleNotFoundError:
        logging.warning(f"Unable to initialize DBeaver connection to DB {name}, config extractor for {plugin_name} not found")
        return


def run_cmd_detached(command):
    kwargs = {
        #'stdout'=subprocess.DEVNULL,
        #'stderr'=subprocess.DEVNULL,
        #'stdin'=subprocess.DEVNULL
    }
    if os.name == 'posix':
        kwargs['start_new_session'] = True
    elif os.name == 'nt':
        kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP | subprocess.DETACHED_PROCESS

    return subprocess.Popen(command, **kwargs)

def run_dbeaver(permanent=False, connect=False, instances=None, print_command=False):
    try:
        dbeaver_path = findapp.findapp(cfg.get("fs.command.dbeaver", "dbeaver-ce"),
                                       app_name="dbeaver",
                                       by_os={'windows': {'binary_name': cfg.get("fs.command.dbeaver-cli", 'dbeaver-cli')}})
    except FileNotFoundError:
        logging.error("DBeaver not found, skipping")
        return

    instances_cfg = cfg.cd("db.instance")
    if instances:
        force_active = True
    else:
        instances = instances_cfg.keys()
        force_active = False

    conargs = []
    for instance in instances:
        try:
            if (dbc := _conarg_lookup(instance)) is not None:
                if force_active or (instance == "work") or dbc.active():
                    args = dbc.conargs()
                    args['folder'] = cfg['fs.project']
                    args['create'] = True
                    if permanent:
                        args['save'] = True
                    if connect:
                        args['connect'] = True
                    conargs.append(args)
        except:
            logging.exception(f"Unable to initialize DBeaver connection to DB {instance}")

    if not conargs:
        logging.error("No active DBeaver connections found or initialized, skipping DBeaver launch")
        return

    cmd = [dbeaver_path]
    for args in conargs:
        cmd.append("-con"),
        cmd.append("|".join([f"{k}={v}" for k, v in args.items()]))

    logging.info(f"Running command {cmd}")

    if print_command:
        from shlex import quote
        print(" ".join([quote(arg) for arg in cmd]))
    else:
        run_cmd_detached(cmd)



"""
This module provides functionality for generating reports from Quarto markdown files within the plpipes framework.

It includes a runner class for executing Quarto rendering actions, along with utilities for handling YAML headers in Quarto files, patching Quarto markdown files with necessary configurations, and managing context changes during execution.
"""

import logging
from plpipes.action.base import Action
from plpipes.action.registry import register_class
from plpipes.config import cfg
import plpipes.filesystem as fs
import subprocess
from pathlib import Path
import yaml
from contextlib import contextmanager
import os
import tempfile
import json
import shutil
import datetime
import friendlydateparser

def _read_yaml_header(fn):
    """
    Reads the YAML header from a Quarto file.

    Args:
        fn (str): The file name of the Quarto file.

    Returns:
        str: The contents of the YAML header if found, otherwise None.

    Raises:
        Exception: If the YAML header is not properly closed.
    """
    with open(fn, "r") as f:
        in_yaml = False
        yaml = []

        for line in f.readlines():
            line = line.rstrip()
            if in_yaml:
                if line == "---":
                    logging.debug(f"YAML header read from Quarto file at {fn}")
                    return "\n".join([*yaml, ""])
                else:
                    yaml.append(line)
            else:
                if line == '---':
                    in_yaml = True
                elif line != '':
                    return None
        else:
            if in_yaml:
                raise Exception("YAML header never closed")
        return None

def _patch_qmd(source, dest, config):
    """
    Patches the Quarto markdown file with required configurations.

    Args:
        source (str): The source Quarto file path.
        dest (str): The destination file path where the patched content will be written.
        config (Path): The configuration file path to be inserted into the Quarto file.
    """
    patch = """```{python}
#| echo: false
import plpipes.action.driver.quarto
plpipes.action.driver.quarto._init_plpipes(""" + repr(str(config.absolute())) + """)
```
"""
    with open(source, "r") as src_f:
        with open(dest, "w") as dest_f:
            state = "start"
            for line in src_f.readlines():
                if state != "copy":
                    stripped = line.rstrip()
                    if state == "start":
                        if stripped == "---":
                            state = "in_yaml"
                        elif stripped != "":
                            dest_f.write(patch)
                            state = "copy"
                    elif state == "in_yaml":
                        if stripped == "---":
                            state = "after_yaml"
                    elif state == "after_yaml":
                        dest_f.write(patch)
                        state = "copy"
                    else:
                        raise Exception("Internal error, unexpected state")
                dest_f.write(line)

def _init_plpipes(config):
    """
    Initializes the PLPIPES framework with the provided configuration.

    Args:
        config (Path): The configuration file path to initialize the framework.
    """
    import plpipes.init
    plpipes.init.init(config_files=[config])

@contextmanager
def _cd(newdir):
    """
    Context manager for changing the current working directory.

    Args:
        newdir (str): The new directory to change to.

    Yields:
        None: The context manager changes to the new directory and reverts to
        the original directory upon exiting the context.
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

class _QuartoRunner(Action):
    """
    Action class for executing Quarto rendering actions.

    Attributes:
        _path (Path): The absolute path to the Quarto markdown file.
    """

    def __init__(self, name, action_cfg):
        """
        Initializes the Quarto runner action.

        Args:
            name (str): The name of the action.
            action_cfg (Config): The configuration settings for the action.
        """
        self._path = Path(action_cfg["files.qmd"]).absolute()
        header = _read_yaml_header(self._path)
        if header is not None:
            action_cfg.merge(yaml.load(header, Loader=yaml.SafeLoader), key="quarto")
        dcfg = action_cfg.cd("dest")
        scfg = action_cfg.cd("quarto.plpipes.dest")
        dcfg.copydefaults(scfg,
                          section='work',
                          dir='',
                          format="html",
                          append_date=False,
                          insert_date=False,
                          date=None)
        default_file = self._path.with_suffix("." + dcfg["format"]).name
        dcfg.copydefaults(scfg, file=default_file)
        super().__init__(name, action_cfg)

    def do_it(self):
        """
        Executes the Quarto rendering action.

        This method handles the preparation of the target file path, date handling,
        and the actual call to the Quarto binary to render the markdown file.
        """
        logging.debug(f"Action config: {self._cfg.to_tree()}")
        rel_path = Path(self._cfg['dest.dir']) / self._cfg['dest.file']

        if self._cfg['dest.insert_date'] or self._cfg['dest.append_date']:
            date_raw = self._cfg['dest.date']
            if date_raw is None:
                date = datetime.datetime.strptime(cfg['run.as_of_date_normalized'], "%Y%m%dT%H%M%SZ0")
            else:
                date = friendlydateparser.parse_datetime(date_raw)
                if date is None:
                    raise ValueError(f"Unable to parse date {date_raw}")

            if self._cfg['dest.insert_date']:
                rel_path = date.strftime(str(rel_path))
                logging.debug(f"Inserting date into target filename ({date} --> {rel_path})")
            else:
                rel_path_parent = rel_path.parent
                rel_path_stem = rel_path.stem
                rel_path_suffix = rel_path.suffix
                rel_path = rel_path_parent / (rel_path_stem + '-' + date.strftime("%Y%m%dT%H%M%SZ0") + rel_path_suffix)
                logging.debug(f"Appending date into target filename ({rel_path})")

        target_path = fs.path(rel_path, section=self._cfg['dest.section'])
        stem = target_path.stem

        with tempfile.TemporaryDirectory() as workdir:
            # workdir = "/home/salva/tmp/report"
            workdir = Path(workdir)
            temp_target_path = workdir / target_path.name
            temp_cfg_path = workdir / "config.json"
            with open(temp_cfg_path, "w") as f:
                json.dump(cfg.to_tree(), f)

            temp_qmd_path = workdir / f"{stem}.qmd"
            _patch_qmd(self._path, temp_qmd_path, temp_cfg_path)

            env = os.environ.copy()
            env['PLPIPES_ROOT_DIR'] = str(fs.path(".", section="root").absolute())

            quarto_binary = cfg.get('bin.quarto', 'quarto')

            with _cd(workdir):
                cmd = [quarto_binary, "render", temp_qmd_path.name,
                       "--output", temp_target_path.name,
                       "--to", self._cfg["dest.format"],
                       "--no-execute-daemon"]

                logging.debug(f"Running quarto: {cmd}, cwd: {os.getcwd()}")
                proc = subprocess.run(cmd, env=env)
                if proc.returncode != 0:
                    raise Exception("Quarto command failed, RC: {proc.returncode}")

                logging.debug("Moving Quarto output files to final destination")
                target_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(temp_target_path,
                             target_path)
                aux_dir_path = Path(f"{stem}_files")
                if aux_dir_path.is_dir():
                    shutil.copytree(aux_dir_path,
                                    target_path.parent / aux_dir_path,
                                    dirs_exist_ok=True)

        logging.info(f"Quarto document available at {target_path}")

register_class("quarto", _QuartoRunner, "qmd")



from plpipes.config import cfg
import plpipes.plugin

import openai
from openai import *
openai.api_key = cfg["cloud.openai.auth.api_key"]

_provider_registry = plpipes.plugin.Registry("openai_provider", "plpipes.cloud.openai.provider.plugin")

_client_registry = {}

def openai_client(name=None):
    if name is None:
        name = "default"

    if name not in _client_registry:
        _client_registry[name] = _init_client(name)
    return _client_registry[name]

def _init_client(name):
    client_cfg = cfg.cd(f"cloud.openai.client.{name}")
    provider_name = client_cfg.get("provider", "openai")
    provider_class = _provider_registry.lookup(provider_name)
    client = provider_class.get_client(name, client_cfg)
    return client



import logging
import sqlalchemy.engine
from urllib.parse import urlunparse, urlparse

from plpipes.database.driver.mysql import MySQLBaseDriver
from plpipes.plugin import plugin

@plugin
class MySQLDriver(MySQLBaseDriver):

    def _default_sqla_driver(self):
        return "mysql"



from plpipes.config import cfg
import plpipes.filesystem as fs
import plpipes.plugin
import logging

_spark_session = None

_driver_class_registry = plpipes.plugin.Registry("spark_plugin", "plpipes.spark.plugin")

def spark_session():
    """Retrieve or create a Spark session.

    This function checks if a Spark session already exists and returns it.
    If not, it initializes a new Spark session based on the configuration
    specified in the 'spark' section of the configuration object.

    Returns:
        Spark session instance.
    """
    global _spark_session
    if _spark_session is None:
        _spark_session = _init_spark_session()
    return _spark_session

def __dir_to_config(*args, url=False, **kwargs):
    """Convert directory path to a string representation.

    This function constructs a file path from given arguments and
    options, resolving the path and converting it to a URL if specified.

    Args:
        *args: Arguments to build the file path.
        url (bool): If True, returns the path as a URL. Defaults to False.
        **kwargs: Additional keyword arguments for path construction.

    Returns:
        str: The constructed file path or URL.
    """
    s = str(fs.path(*args, **kwargs).resolve()).replace("\\", "/")
    if url:
        return f'file://{s}'
    return s

def _init_spark_session():
    """Initialize a new Spark session.

    This function retrieves the Spark configuration from the global
    configuration object, looks up the specified driver class,
    and initializes a Spark session. It uses the configuration
    provided for Spark settings.

    Returns:
        Spark session instance.
    """
    ssc = cfg.cd("spark")
    driver_name = ssc.get("driver", "embedded")
    driver_class = _driver_class_registry.lookup(driver_name)
    driver = driver_class()
    return driver.init_spark_session(ssc)



from plpipes.config import cfg

import sys
import os
import logging
import pathlib
import datetime
import friendlydateparser

_initialized = False

_config0 = {'db': {'instance': {'work': {},
                                'input': {},
                                'output': {}}},
            'env': os.environ.get("PLPIPES_ENV",
                                  "dev"),
            'run': {'as_of_date': 'now'},
            'logging': {'level': os.environ.get("PLPIPES_LOGLEVEL",
                                                "INFO"),
                        'level_file': os.environ.get("PLPIPES_LOGLEVEL_FILE",
                                                     "INFO")}}

def init(*configs, config_files=[]):
    """Initializes the PLPipes configuration.

    This function merges provided configurations, command line arguments, and environment variables
    to set up the application. It also initializes logging and file system paths.

    Args:
        *configs: Additional configuration dictionaries.
        config_files: List of custom configuration files to be merged.

    Returns:
        bool: True if initialization is successful.
    """
    from pathlib import Path

    global _initialized
    if _initialized:
        logging.warning("Reinitialicing PLPipes")

    # frame 0: command line arguments
    # frame 1: custom configuration files
    # frame 2: standard configuration files

    cfg.merge(_config0, frame=2)

    for config in configs:
        for k, v in config.items():
            cfg.merge(v, key=k, frame=0)

    prog = Path(sys.argv[0])
    default_stem = str(prog.stem)

    # update cfg to get the log levels (normal and file)
    for fn in config_files:
        cfg.merge_file(fn, frame=1)

    list_configuration_files_not_found = []

    global_cfg_dir = Path.home() / ".config/plpipes"
    for suffix in ("", "-secrets"):
        for ext in ("json", "yml", "yaml"):
            path = global_cfg_dir / f"plpipes{suffix}.{ext}"
            if path.exists():
                cfg.merge_file(path, frame=2)
            else:
                list_configuration_files_not_found.append(path)

    for dir_key in (False, True):
        for stem_key in (False, True):
            for secrets_part in ("", "-secrets"):
                for env_key in (False, True):
                    for ext in ("json", "yml", "yaml"):
                        # The following values can be changed as
                        # config files are read, so they are
                        # recalculated every time:

                        env         = cfg.get('env', 'dev')
                        stem        = cfg.get('fs.stem', default_stem)
                        root_dir    = Path(cfg.get('fs.root'   , prog.parent.parent.absolute()))
                        config_dir  = Path(cfg.get('fs.config' , root_dir / "config"))
                        default_dir = Path(cfg.get('fs.default', root_dir / "default"))

                        env_part  = f"-{env}"  if env_key  else ""
                        stem_part = stem       if stem_key else "common"
                        dir       = config_dir if dir_key  else default_dir
                        path      = dir / f"{stem_part}{secrets_part}{env_part}.{ext}"
                        if path.exists():
                            cfg.merge_file(path, frame=2)
                        else:
                            list_configuration_files_not_found.append(path)

    # set the root log level as NOTSET, which is the deepest level; it's like
    # this because all the other handlers, even if they have ther own levels,
    # will not log anything if the root level is higher than their level
    logging.getLogger().setLevel("NOTSET")

    cfg.squash_frames()

    cfg.setdefault('fs.stem', default_stem)

    # calculate configuration for file system paths and set it
    root_dir = Path(cfg.setdefault('fs.root', prog.parent.parent.absolute()))
    for e in ('bin', 'lib', 'config', 'default',
              'input', 'output', 'work', 'actions',
              'resources'):
        cfg.setdefault("fs." + e, root_dir / e)

    init_run_as_of_date()

    _log_setup()

    logging.debug(f"List of configuration files not found: {list_configuration_files_not_found}")

    logging.debug(f"Configuration: {repr(cfg.to_tree())}")

    _initialized = True

    return True

def init_run_as_of_date():
    """Initializes the 'as_of_date' configuration entry.

    It sets the 'run.as_of_date_normalized' entry in the configuration
    based on the specified or default 'run.as_of_date'.

    Returns:
        None
    """
    date = cfg.setdefault('run.as_of_date', 'now')
    as_of_date = friendlydateparser.parse_datetime(date)
    as_of_date = as_of_date.astimezone(datetime.timezone.utc)
    cfg['run.as_of_date_normalized'] = as_of_date.strftime("%Y%m%dT%H%M%SZ0")

def _log_setup():
    """This function sets up the logging system. It is called by init().

    Args:
        None.
    Returns:
        None
    """

    # get the logger
    logger = logging.getLogger()
    formatter = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s:%(message)s")

    # if logging.log_to_file is True, then we set up a file handler
    if cfg.get("logging.log_to_file", True):
        dir = cfg.get("logging.log_dir", "logs")
        dir = pathlib.Path(cfg["fs.root"]) / dir
        dir.mkdir(exist_ok=True, parents=True)
        ts = datetime.datetime.utcnow().strftime('%y%m%dT%H%M%SZ')
        name = f"log-{ts}.txt"

        last = dir / "last-log.txt"
        try:
            if last.exists():
                last.unlink()
            (dir / "last-log.txt").symlink_to(name)
        except:
            import platform
            if platform.system() != 'Windows':
                logging.warn(f"Unable to create link for {last}", exc_info=True)

        fh = logging.FileHandler(str(dir / name))
        fh.setLevel(cfg["logging.level_file"].upper())
        fh.setFormatter(formatter)

    # we set up a console handler
    ch = logging.StreamHandler()
    ch.setLevel(cfg["logging.level"].upper())

    if ch.stream.isatty():
        import colorlog
        ch.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(asctime)s:%(levelname)s:%(name)s:%(reset)s%(white)s%(message)s",
                                                  log_colors={'DEBUG': 'cyan',
                                                              'INFO': 'green',
                                                              'WARNING': 'yellow',
                                                              'ERROR': 'red',
                                                              'CRITICAL': 'red,bg_white'}))
    else:
        ch.setFormatter(formatter)

    # we force the handlers as ch and fh
    if cfg.get("logging.log_to_file", True):
        logger.handlers = [ch, fh]
    else:
        logger.handlers = [ch]

    # we do not allow propagations to other handlers
    logger.propagate = False




from plpipes.plugin import plugin
from plpipes.cloud.azure.auth.base import AuthenticatorBase
from azure.identity import InteractiveBrowserCredential, TokenCachePersistenceOptions, AuthenticationRecord
import logging

from plpipes.exceptions import AuthenticationError


@plugin
class IBCAuthenticator(AuthenticatorBase):

    def _authenticate(self):
        ar_fn = self._credentials_cache_filename()
        try:
            with open(ar_fn, "r") as f:
                ar = AuthenticationRecord.deserialize(f.read())
        except:
            logging.debug(f"Couldn't load authentication record for {self._account_name} from {ar_fn}")
            ar = None

        authentication_callback_port = self._cfg.setdefault("authentication_callback_port", 8082)
        redirect_uri = f"http://localhost:{authentication_callback_port}"

        allow_unencrypted_storage = self._cfg.setdefault("allow_unencrypted_storage", False)
        cache_persistence_options = TokenCachePersistenceOptions(allow_unencrypted_storage=allow_unencrypted_storage)
        expected_user = self._cfg.get("username")

        cred = InteractiveBrowserCredential(tenant_id=self._cfg.get("tenant_id", "common"),
                                            client_id=self._cfg["client_id"],
                                            client_credential=self._cfg.get("client_secret"),
                                            prompt=self._cfg.get("prompt", "login"),
                                            login_hint=expected_user,
                                            redirect_uri=redirect_uri,
                                            cache_persistence_options=cache_persistence_options,
                                            authentication_record=ar)
        if "scopes" in self._cfg:
            scopes = self._cfg["scopes"]
            if isinstance(scopes, str):
                scopes = scopes.split(" ")

            logging.debug(f"Calling authenticate(scopes={scopes})")
            ar = cred.authenticate(scopes=scopes)

            if expected_user not in (None, ar.username):
                raise AuthenticationError(f"Authenticating as user {expected_user} expected but {ar.username} found!")
            try:
                logging.debug(f"Saving authentication record to {ar_fn}")
                ar_fn.parent.mkdir(parents=True, exist_ok=True)
                with open(ar_fn, "w") as f:
                    f.write(ar.serialize())
            except:
                logging.warning(f"Unable to save authentication record for {self._account_name} at {ar_fn}", exc_info=True)
        else:
            logging.warning(f"'cloud.azure.graph.{self._account_name}.scopes' not configured, credentials for {self._account_name} are not going to be cached!")

        return cred




"""
This module provides SQLAlchemy extensions for performing tasks not natively supported by the SQLAlchemy ORM.
It includes constructs for creating and dropping tables and views, inserting data from queries,
and handling subqueries.
"""

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import ClauseElement, Executable, FromClause
import sqlalchemy.sql
import logging

class _CreateSomethingAs(Executable, ClauseElement):
    """
    Base class for creating a table or view from a select statement.
    """
    def __init__(self, table_or_view, table_name, select, if_not_exists=False):
        """
        Initializes the creation of a table or view.

        :param table_or_view: Type of object being created ('TABLE' or 'VIEW').
        :param table_name: Name of the table or view to be created.
        :param select: Select statement to create the object from.
        :param if_not_exists: Whether to create the object only if it does not exist.
        """
        self._table_name = table_name
        self._select = select
        self._if_not_exists = if_not_exists
        self._table_or_view = table_or_view

class CreateTableAs(_CreateSomethingAs):
    """
    Class for creating a table from a select statement.
    """
    inherit_cache = True

    def __init__(self, *args, **kwargs):
        """
        Initializes the creation of a table.
        """
        super().__init__("TABLE", *args, **kwargs)

class CreateViewAs(_CreateSomethingAs):
    """
    Class for creating a view from a select statement.
    """
    inherit_cache = True

    def __init__(self, *args, **kwargs):
        """
        Initializes the creation of a view.
        """
        super().__init__("VIEW", *args, **kwargs)

@compiles(_CreateSomethingAs)
def _create_something_as(element, compiler, **kw):
    """
    Compiles the creation of a table or view into a SQL statement.

    :param element: The element being compiled.
    :param compiler: The SQL compiler.
    :param kw: Additional options for compilation.
    :return: A SQL statement for creating the specified object.
    """
    table_name = compiler.preparer.quote_identifier(element._table_name)
    select = compiler.process(element._select, **kw)
    if element._if_not_exists:
        sql = f"CREATE {element._table_or_view} IF NOT EXISTS {table_name} AS {select}"
    else:
        sql = f"CREATE {element._table_or_view} {table_name} AS {select}"
    logging.debug(f"SQL code: {sql}")
    return sql

class InsertIntoTableFromQuery(Executable, ClauseElement):
    """
    Class for inserting data into a table from a select statement.
    """
    inherit_cache = True

    def __init__(self, table_name, select):
        """
        Initializes the insert operation.

        :param table_name: Name of the table to insert into.
        :param select: Select statement providing data to insert.
        """
        self._table_name = table_name
        self._select = select

@compiles(InsertIntoTableFromQuery)
def _insert_into_table_from_query(element, compiler, **kw):
    """
    Compiles the insert operation into a SQL statement.

    :param element: The element being compiled.
    :param compiler: The SQL compiler.
    :param kw: Additional options for compilation.
    :return: A SQL statement for inserting data into the specified table.
    """
    table_name = compiler.preparer.quote_identifier(element._table_name)
    select = compiler.process(element._select, **kw)
    sql = f"INSERT INTO {table_name} {select}"
    logging.debug(f"SQL code: {sql}")
    return sql

class _DropSomething(Executable, ClauseElement):
    """
    Base class for dropping a table or view.
    """
    def __init__(self, table_or_view, table_name, if_exists=False):
        """
        Initializes the drop operation.

        :param table_or_view: Type of object being dropped ('TABLE' or 'VIEW').
        :param table_name: Name of the table or view to be dropped.
        :param if_exists: Whether to drop the object only if it exists.
        """
        self._table_name = table_name
        self._if_exists = if_exists
        self._table_or_view = table_or_view

class DropTable(_DropSomething):
    """
    Class for dropping a table.
    """
    inherit_cache = False

    def __init__(self, *args, **kwargs):
        """
        Initializes the drop operation for a table.
        """
        super().__init__("TABLE", *args, **kwargs)

class DropView(_DropSomething):
    """
    Class for dropping a view.
    """
    inherit_cache = False

    def __init__(self, *args, **kwargs):
        """
        Initializes the drop operation for a view.
        """
        super().__init__("VIEW", *args, **kwargs)

@compiles(_DropSomething)
def _drop_something(element, compiler, **kwargs):
    """
    Compiles the drop operation into a SQL statement.

    :param element: The element being compiled.
    :param compiler: The SQL compiler.
    :param kwargs: Additional options for compilation.
    :return: A SQL statement for dropping the specified object.
    """
    if element._if_exists:
        sql = f"DROP {element._table_or_view} IF EXISTS {element._table_name}"
    else:
        sql = f"DROP {element._table_or_view} {element._table_name}"
    logging.debug(f"SQL code: {sql}")
    return sql

class AsSubquery(FromClause):
    """
    Class for handling subqueries.
    """
    inherit_cache = False

    def __init__(self, txt):
        """
        Initializes a subquery.

        :param txt: The text of the subquery.
        """
        self._txt = txt

@compiles(AsSubquery)
def _as_subquery(element, compiler, **kwargs):
    """
    Compiles the subquery into a SQL statement.

    :param element: The element being compiled.
    :param compiler: The SQL compiler.
    :param kwargs: Additional options for compilation.
    :return: A SQL representation of the subquery.
    """
    txt = compiler.process(element._txt)
    return f"({txt})"

def Wrap(str_or_something):
    """
    Wraps a string or other expression into a SQLAlchemy text object.

    :param str_or_something: The string or expression to wrap.
    :return: A SQLAlchemy text object or the original expression.
    """
    if isinstance(str_or_something, str):
        return sqlalchemy.sql.text(str_or_something)
    return str_or_something



import plpipes.plugin

class ConArg(plpipes.plugin.Plugin):
    _con_args = ['name', 'driver', 'url', 'host', 'port', 'server', 'database', 'user', 'password', 'auth']
    _con_boolean_args = ['save', 'connect', 'openConsole', 'create']

    #def _init_plugin(self, name, drv_cfg):
    #    pass

    def __init__(self, name, db_drv):
        self.name = name
        self._db_drv = db_drv
        self._cfg = db_drv.config()
        super().__init__()

    def active(self):
        return True

    def conargs(self):
        args = {}
        for arg in self._con_args:
            value = getattr(self, arg, None)
            if value is not None:
                if "|" in str(value):
                    raise ValueError(f"Value for {arg} for database instance {self.name} contains a pipe character, which is not allowed")
                args[arg] = value

        for arg in self._con_boolean_args:
            value = getattr(self, arg, None)
            if value is not None:
                args[arg] = "true" if value else "false"
        return args




from plpipes.plugin import plugin
from plpipes.tool.dbeaver.conarg.mysql import MySQLBaseConArg

@plugin
class MySQLConArg(MySQLBaseConArg):

    def __init__(self, name, db_drv):
        super().__init__(name, db_drv)
        self.driver = "mysql"



"""
plpipes.autoaction

This module is a "magic" module that enables Python actions to be executed as if they were standard scripts. It should be imported at the beginning of any action script, as it performs necessary initialization and variable injection.

When imported, this module checks if the plpipes environment is initialized. If it is not, it sets up the environment and injects variables into the calling code's global namespace. This includes variables defined by PLPipes for actions, making them readily available for the action script without requiring explicit definitions.

Usage:
- Import this module at the very start of your action script to ensure proper initialization and variable accessibility.
- The module will automatically handle the necessary setup and variable injection.

Example:

```python
import plpipes.autoaction
# Your action script logic follows...
```

Exceptions:
- If the module is not the first import in the script, an exception will be raised indicating that the plpipes.autoaction must be the first sentence in the action script.

"""

import plpipes.init

if not plpipes.init._initialized:
    from plpipes.runner import simple_init
    from plpipes.action.driver.simple import _action_namespace_setup
    import inspect
    import logging
    import sys

    argv = sys.argv
    cmd = argv[0] if argv else ''
    # Hack for VSCode/ipykernel not setting the correct command line arguments
    if cmd.endswith('ipykernel_launcher.py'):
        argv = ['bin/run.py']

    simple_init(argv)

    def get_main_frame():
        stack = inspect.stack()
        for frame_info in stack:
            if frame_info.function == '<module>':
                if frame_info.frame.f_globals['__name__'] == '__main__':
                    return frame_info.frame

        raise Exception("Unable to locate main frame for injecting globals!")

    main = get_main_frame()
    for k, v in main.f_globals.items():
        if k.startswith("__"):
            continue
        logging.warning(f"Importing plpipes.autoaction should be the first sentence in the action script. Found `{k}` already declared.")

    logging.info(f"Injecting variables into action namespace.")

    for k, v in _action_namespace_setup().items():
        main.f_globals[k] = v



from plpipes.plugin import plugin
from plpipes.cloud.azure.auth.base import AuthenticatorBase
from azure.identity import ClientSecretCredential
import logging

from plpipes.exceptions import AuthenticationError


@plugin
class MIAuthenticator(AuthenticatorBase):
    def _authenticate(self):
        return ManagedIdentityCredential()




import logging
from plpipes.database.driver.odbc import ODBCDriver

class SQLServerDriver(ODBCDriver):
    def __init__(self, name, drv_cfg):
        if "connection_string" not in drv_cfg:
            driver = drv_cfg.setdefault('driver_odbc', '{ODBC Driver 18 for SQL Server}')
            port = drv_cfg.get('port')
            low_level_proto = drv_cfg.get('low_level_proto', 'tcp')
            encrypt = 'yes' if drv_cfg.setdefault('encrypt', True) else 'no'
            trust_server_certificate = 'yes' if drv_cfg.setdefault('trust_server_certificate', True) else 'no'
            trusted_connection = 'yes' if drv_cfg.setdefault('trusted_connection', True) else 'no'
            timeout = drv_cfg.setdefault('timeout', 60)
            server = drv_cfg['server']
            database = drv_cfg['database']
            uid = drv_cfg.get('user')
            password = drv_cfg.get('password')

            cs = {'Driver': driver,
                  'Database': database,
                  'TrustServerCertificate': trust_server_certificate,
                  'Connection Timeout': str(timeout),
                  'TrustedConnection': trusted_connection}

            if "\\" not in server:
                server = f"{low_level_proto}:{server}"
                if port is not None:
                    server += f",{port}"
                cs['Encrypt'] = encrypt
            cs['Server'] = server

            if uid is not None:
                cs['Uid'] = uid
            if password is not None:
                cs['Pwd'] = password

            cs = ";".join([f"{k}={v}" for k,v in cs.items()])
            # logging.info(f"Connection string: {cs}")

            drv_cfg['connection_string'] = cs
        drv_cfg.setdefault('sql_alchemy_driver', 'mssql+pyodbc')
        super().__init__(name, drv_cfg)



import logging

from plpipes.database.backend import Backend
from plpipes.plugin import plugin
from plpipes.spark import spark_session

class SparkBackendBase(Backend):
    def query(self, txn, sql, parameters, kws):
        df = txn._conn.sql(sql, args=parameters, **kws)
        return self._coerce_output_df(df)

    def read_table(self, txn, table_name, kws):
        df = txn._conn.read.table(table_name)
        return self._coerce_output_df(df)

    def _coerce_output_df(self, df):
        return df

    def _create_table_from_spark(self, txn, table_name, df, parameters, if_exists, kws):
        logging.debug(f"Creating table {table_name} from spark dataframe, if_exists={if_exists}")
        if if_exists == 'replace':
            txn.drop_table(table_name, True)
            mode = 'overwrite'
        elif if_exists == 'append':
            mode = 'append'
        elif if_exists == 'ignore':
            mode = 'ignore'
        else:
            raise ValueError(f"Bad value {if_exists} for if_exists")

        logging.debug(f"Writing dataframe to table {table_name} with mode {mode}")
        df.write.mode(mode).saveAsTable(table_name)



from plpipes.database.backend.spark import SparkBackendBase
from plpipes.plugin import plugin

from pyspark.sql.dataframe import DataFrame

@plugin
class SparkBackend(SparkBackendBase):

    def register_handlers(self, handlers):
        handlers['create_table'].register(DataFrame, self._create_table_from_spark)




import logging
import os

from plpipes.database.driver.sqlite import SQLiteDriver
from plpipes.plugin import plugin

from sqlalchemy import event

@plugin
class SpatialiteDriver(SQLiteDriver):

    _default_backend_name = "geopandas"

    def __init__(self, name, drv_cfg):
        os.environ["SPATIALITE_SECURITY"] = "relaxed"

        super().__init__(name, drv_cfg)

        @event.listens_for(self._engine, "connect")
        def init_spatialite_connection(conn, cr):
            logging.debug("Initializing SQLite connection, loading mod_spatialite")
            conn.enable_load_extension(True)
            conn.execute("select load_extension('mod_spatialite')")

        self._engine.execute("select InitSpatialMetaData()")



from plpipes.config import cfg
import logging
import plpipes.plugin
import plpipes.database.driver
import plpipes.database.driver.transaction

_driver_registry = plpipes.plugin.Registry("db_driver", "plpipes.database.driver.plugin")
_db_registry = {}

def lookup(db=None):
    """
    Lookup the database driver instance for the specified database.

    Args:
        db (str, optional): The name of the database instance to look up. Defaults to "work".

    Returns:
        driver: The database driver instance.
    """
    if db is None:
        db = "work"
    if db not in _db_registry:
        _db_registry[db] = _init_driver(db)
    return _db_registry[db]

def _init_driver(name):
    """
    Initialize the database driver for the specified instance name.

    Args:
        name (str): The name of the database instance to initialize.

    Returns:
        driver: The initialized database driver.
    """
    drv_cfg = cfg.cd(f"db.instance.{name}")
    driver_name = drv_cfg.setdefault("driver", "sqlite")
    logging.debug(f"Initializing database instance {name} using driver {driver_name}")
    driver = _driver_registry.lookup(driver_name)
    return driver(name, drv_cfg)

def begin(db=None):
    """
    Begin a transaction on the specified database instance.

    Args:
        db (str, optional): The name of the database instance to begin a transaction. Defaults to "work".

    Returns:
        Transaction: A transaction object for the database.
    """
    return lookup(db).begin()

class _TxnWrapper:
    """
    A wrapper class to handle transaction context management.
    """
    def __init__(self, txn):
        self._txn = txn

    def __enter__(self):
        return self._txn

    def __exit__(self, a, b, c):
        return False

def _begin_or_pass_through(db_or_txn):
    """
    Begin a transaction or pass through to the existing transaction.

    Args:
        db_or_txn: Either a database name or an existing transaction.

    Returns:
        TransactionWrapper: A context manager for the transaction.
    """
    if isinstance(db_or_txn, plpipes.database.driver.transaction.Transaction):
        return _TxnWrapper(db_or_txn)
    else:
        return lookup(db_or_txn).begin()

def query_first(sql, parameters=None, db=None, backend=None, **kws):
    """
    Execute a SQL query and return the first result.

    Args:
        sql (str): The SQL query to execute.
        parameters (dict, optional): The parameters for the SQL query.
        db (str, optional): The database instance to use.
        backend (str, optional): The backend to use.

    Returns:
        object: The first result of the query.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.query_first(sql, parameters, backend, **kws)

def query_first_value(sql, parameters=None, db=None, backend="tuple", **kws):
    """
    Execute a SQL query and return the first value from the result.

    Args:
        sql (str): The SQL query to execute.
        parameters (dict, optional): The parameters for the SQL query.
        db (str, optional): The database instance to use.
        backend (str, optional): The backend to use. Defaults to "tuple".

    Returns:
        object: The first value from the query result.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.query_first_value(sql, parameters, backend, **kws)

def query(sql, parameters=None, db=None, backend=None, **kws):
    """
    Execute a SQL query and return the results.

    Args:
        sql (str): The SQL query to execute.
        parameters (dict, optional): The parameters for the SQL query.
        db (str, optional): The database instance to use.
        backend (str, optional): The backend to use.

    Returns:
        DataFrame: The results of the query as a DataFrame.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.query(sql, parameters, backend, **kws)

def execute(sql, parameters=None, db=None):
    """
    Execute a SQL command that does not return a result set.

    Args:
        sql (str): The SQL command to execute.
        parameters (dict, optional): The parameters for the SQL command.
        db (str, optional): The database instance to use.

    Returns:
        None
    """
    with _begin_or_pass_through(db) as txn:
        return txn.execute(sql, parameters)

def create_table(table_name, sql_or_df, parameters=None, db=None, if_exists="replace", **kws):
    """
    Create a new table in the database from a DataFrame, SQL command, or SQLAlchemy select object.

    Args:
        table_name (str): The name of the table to create.
        sql_or_df (DataFrame, str, or SQLAlchemy select object): The DataFrame, SQL command, or SQLAlchemy select object for creating the table.
        parameters (dict, optional): The parameters for creating the table.
        db (str, optional): The database instance to use.
        if_exists (str, optional): What to do if the table already exists. Defaults to "replace".
        **kws: Additional keyword arguments.

    Returns:
        None
    """
    logging.debug(f"create table {table_name}")
    with _begin_or_pass_through(db) as txn:
        return txn.create_table(table_name, sql_or_df, parameters, if_exists, **kws)

def create_view(view_name, sql, parameters=None, db=None, if_exists="replace", **kws):
    """
    Create a new view in the database.

    Args:
        view_name (str): The name of the view to create.
        sql (str): The SQL command for creating the view.
        parameters (dict, optional): The parameters for creating the view.
        db (str, optional): The database instance to use.
        if_exists (str, optional): What to do if the view already exists. Defaults to "replace".
        **kws: Additional keyword arguments.

    Returns:
        None
    """
    with _begin_or_pass_through(db) as txn:
        return txn.create_view(view_name, sql, parameters, if_exists, **kws)

def read_table(table_name, db=None, backend=None, **kws):
    """
    Read the contents of a table and return it as a data frame.

    Args:
        table_name (str): The name of the table to read.
        db (str, optional): The database instance to use.
        backend (str, optional): The backend to use.
        **kws: Additional keyword arguments.

    Returns:
        DataFrame: The contents of the table as a DataFrame.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.read_table(table_name, backend, **kws)

def execute_script(sql_script, db=None):
    """
    Execute a sequence of SQL commands.

    Args:
        sql_script (str): The SQL script to execute.
        db (str, optional): The database instance to use.

    Returns:
        None
    """
    with _begin_or_pass_through(db) as txn:
        return txn.execute_script(sql_script)

def list_tables(db=None):
    """
    List all tables in the specified database.

    Args:
        db (str, optional): The database instance to use.

    Returns:
        list: A list of table names in the database.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.list_tables()

def list_views(db=None):
    """
    List all views in the specified database.

    Args:
        db (str, optional): The database instance to use.

    Returns:
        list: A list of view names in the database.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.list_views()

def table_exists_p(table_name, db=None):
    """
    Check if a table exists in the specified database.

    Args:
        table_name (str): The name of the table to check.
        db (str, optional): The database instance to use.

    Returns:
        bool: True if the table exists, False otherwise.
    """
    with _begin_or_pass_through(db) as txn:
        return txn.table_exists_p(table_name)

def drop_table(table_name, db=None, only_if_exists=False):
    """
    Drop a table from the database.

    Args:
        table_name (str): The name of the table to drop.
        db (str, optional): The database instance to use.
        only_if_exists (bool, optional): If True, do not raise an error if the table does not exist.

    Returns:
        None
    """
    with _begin_or_pass_through(db) as txn:
        return txn.drop_table(table_name, only_if_exists)

def query_chunked(sql, parameters=None, db=None, backend=None, **kws):
    """
    Execute a SQL query and yield results in chunks.

    Args:
        sql (str): The SQL query to execute.
        parameters (dict, optional): The parameters for the SQL query.
        db (str, optional): The database instance to use.
        backend (str, optional): The backend to use.
        **kws: Additional keyword arguments.

    Yields:
        DataFrame: Each chunk of the results as a DataFrame.
    """
    with _begin_or_pass_through(db) as txn:
        for df in txn.query_chunked(sql, parameters, backend, **kws):
            yield df

def query_group(sql, parameters=None, db=None, by=None, backend=None, **kws):
    """
    Execute a SQL query and yield results grouped by specified criteria.

    Args:
        sql (str): The SQL query to execute.
        parameters (dict, optional): The parameters for the SQL query.
        db (str, optional): The database instance to use.
        by (str, optional): The column or columns to group by.
        backend (str, optional): The backend to use.
        **kws: Additional keyword arguments.

    Yields:
        DataFrame: Each group of results as a DataFrame.
    """
    with _begin_or_pass_through(db) as txn:
        for df in txn.query_group(sql, parameters, by, backend, **kws):
            yield df

def copy_table(from_table_name, to_table_name=None,
               from_db=None, to_db=None, db=None,
               if_exists="replace", **kws):
    """
    Copy a table from one database instance to another.

    Args:
        from_table_name (str): The name of the source table to copy.
        to_table_name (str, optional): The name of the destination table. Defaults to the source table name.
        from_db (str, optional): The source database instance to copy from.
        to_db (str, optional): The destination database instance to copy to.
        db (str, optional): The current database instance.
        if_exists (str, optional): What to do if the destination table already exists. Defaults to "replace".
        **kws: Additional keyword arguments.

    Returns:
        None
    """
    if to_table_name is None:
        to_table_name = from_table_name.split(".")[-1]

    if from_db is None:
        from_db = db
    if to_db is None:
        to_db = db

    with _begin_or_pass_through(from_db) as from_txn:
        if from_db == to_db:
            logging.debug(f"copy table {from_table_name} as {to_table_name} inside db {from_txn.db_name()}")
            from_txn.copy_table(from_table_name, to_table_name, if_exists=if_exists, **kws)
        else:
            with _begin_or_pass_through(to_db) as to_txn:
                logging.debug(f"copy table {from_table_name} from db {from_txn.db_name()} as {to_table_name} in db {to_txn.db_name()}")
                if if_exists == "replace":
                    to_txn.drop_table(to_table_name)
                    if_exists="append"
                elif if_exists == "drop_rows":
                    to_txn.drop_table_rows(to_table_name)
                    if_exists="append"
                first = True
                for df in from_txn.read_table_chunked(from_table_name, **kws):
                    if first:
                        to_txn.create_table(to_table_name, df, if_exists=if_exists)
                        first = False
                    else:
                        to_txn.create_table(to_table_name, df, if_exists="append")

_key_dir_unpacked = {
    '>' : (True , True ), # Ascending, Strict
    '>=': (True , False),
    '<' : (False, True ),
    '<=': (False, False)
}

def update_table(from_table_name, to_table_name=None,
                 from_db=None, to_db=None, db=None,
                 key=None, key_dir=">=", **kws):
    """
    Update a table in the destination database from the source table.

    Args:
        from_table_name (str): The name of the source table to update from.
        to_table_name (str, optional): The name of the destination table to update. Defaults to the source table name.
        from_db (str, optional): The source database instance.
        to_db (str, optional): The destination database instance.
        db (str, optional): The current database instance.
        key (str): The key column used to identify new rows.
        key_dir (str, optional): The direction for the key comparison. Defaults to ">=".
        **kws: Additional keyword arguments.

    Returns:
        None
    """
    if to_table_name is None:
        to_table_name = from_table_name

    if from_db is None:
        from_db = db
    if to_db is None:
        to_db = db

    try:
        ascending, strict = _key_dir_unpacked[key_dir]
    except KeyError:
        raise ValueError(f"Invalid key_dir value {key_dir}")

    with _begin_or_pass_through(from_db) as from_txn:
        with _begin_or_pass_through(to_db) as to_txn:
            logging.debug(f"Updating table {from_table_name} from db {from_txn.db_name()} as {to_table_name} in db {to_txn.db_name()}")

            if to_txn.driver()._engine.dialect.has_table(to_txn._conn, to_table_name):
                count = to_txn.query_first_value(f"select count(*) from (select {key} from {to_table_name} limit 1) as t")
                if count > 0:
                    top_func = "max" if ascending else "min"
                    # FIXME: escape key identifier properly!
                    top = to_txn.query_first_value(f"select {top_func}({key}) from {to_table_name}")
                    if not strict:
                        # we don't know whether we already have all the rows with key=top, so we have to also update those!
                        to_txn.execute(f"delete from {to_table_name} where {key} = :top", parameters={'top': top})
                    for df in from_txn.query_chunked(f"select * from {from_table_name} where {key} {key_dir} :top",
                                                     parameters={'top': top}):
                        to_txn.create_table(to_table_name, df, if_exists="append")
                    return
            # No table, or table is empty

            for df in from_txn.read_table_chunked(from_table_name):
                to_txn.create_table(to_table_name, df, if_exists="append")

def engine(db=None):
    """
    Retrieve the database engine for the specified database instance.

    Args:
        db (str, optional): The name of the database instance to retrieve the engine for. Defaults to "work".

    Returns:
        engine: The database engine object.
    """
    return lookup(db).engine()

def load_backend(name, db=None):
    """
    Load a specific backend into the database driver.

    Args:
        name (str): The name of the backend to load.
        db (str, optional): The database instance to load the backend into. Defaults to "work".

    Returns:
        None
    """
    lookup(db).load_backend(name)



# Module for autoconfiguring DBeaver

import sys
from plpipes.tool.dbeaver import run_dbeaver
from plpipes.runner import arg_parser, parse_args_and_init

def run(argv):
    ap = arg_parser(prog="plpipes.tool.dbeaver",
                    description='PLPipes DBeaver autoconfiguration tool',
                    epilog="""This tool is used to automate
                              the configuration of DBeaver with database
                              connections defined in your project.  The
                              database connection details are read from
                              the project-specific configuration files
                              managed by the plpipes framework.""")

    ap.add_argument('-p', '--permanent', '--save', action='store_true',
                    help='Make the connections persistent in DBeaver')
    ap.add_argument('-o', '--connect', '--open', action='store_true',
                    help='Connect immediately to the databases after setting them up')
    ap.add_argument('-i', '--instance', type=str, action='append',
                    help='Initialize only the connection(s) with the given name. Can be used multiple times.')
    ap.add_argument('-q', '--command', action='store_true',
                    help='Print the command that would be used to launch DBeaver and exit')
    opts = parse_args_and_init(ap, argv)

    # Launch DBeaver with the specified options
    run_dbeaver(permanent=opts.permanent, connect=opts.connect, instances=opts.instance, print_command=opts.command)

if __name__ == "__main__":
    run(["plpipes-tool-dbeaver", *sys.argv[1:]])



import sqlite_vss

from plpipes.plugin import plugin, Plugin
from sqlalchemy import event

@plugin
class VssExtension(Plugin):

    def __init__(self, driver, extension_name, drv_cfg):

        @event.listens_for(driver._engine, "connect")
        def init_vss(conn, cr):
            conn.enable_load_extension(True)
            sqlite_vss.load(conn)




import logging

from plpipes.database.driver import Driver
from plpipes.database.driver.transaction import Transaction
from plpipes.plugin import plugin

from influxdb import DataFrameClient
from contextlib import contextmanager

def _check_backend(name):
    assert name in (None, "pandas", "multipandas", "autopandas")

DEFAULT_CHUNKSIZE = 5000

@plugin
class InfluxDB1BDriver(Driver):

    def __init__(self, name, drv_cfg):

        super().__init__(name, drv_cfg)

        host       = drv_cfg.setdefault('host', 'localhost')
        port       = drv_cfg.setdefault('port', '8086')
        database   = drv_cfg.get('database')
        username   = drv_cfg.get('user')
        password   = drv_cfg.get('password')
        ssl        = drv_cfg.setdefault('ssl', True)
        verify_ssl = drv_cfg.setdefault('verify_ssl', True)

        client = DataFrameClient(host=host, port=port, database=database,
                                 username=username, password=password,
                                 ssl=ssl, verify_ssl=verify_ssl)

        self._influxdb_client =  client

    def _backend(self, name):
        raise NotImplementedError(f"Method _backend not implemented (it shouldn't be needed anyway with this driver).")

    @contextmanager
    def begin(self):
        yield Transaction(self, None)

    def _make_result(self, r, backend):
        _check_backend(backend)
        if backend == "multipandas":
            return r
        elif len(r) == 1:
          return list(r.values())[0] # pandas & autopandas
        elif backend == "autopandas":
            return r
        else:
            raise ValueError("pandas backend selected but the response contains more than one dataset (consider using multipandas or autopandas backends instead)")

    def _query(self, txn, query_text, parameters, backend, kws):
        kws.setdefault("dropna", False)
        r = self._influxdb_client.query(query_text, params=parameters, **kws)
        return self._make_result(r, backend)

    def _select_all_from_table_query(self, table_name):
        table_name = table_name.replace('"', '\\"')
        q = f'select * from "{table_name}"'
        logging.debug(f"read table query: {q}")
        return q

    def _read_table(self, txn, table_name, backend, kws):
        return self._query(txn,
                           self._select_all_from_table_query(table_name),
                           None, backend, kws)

    def engine(self):
        return self._influxdb_client

    def _query_chunked(self, txn, query_text, parameters, backend, kws):
        kws.setdefault("dropna", False)
        # FIXME: support for chunked queries is broken in the latest
        # versions of the client module, so we just don't do them!
        #
        # for r in self._influxdb_client.query(query_text, params=parameters,
        #                                      chunked=True, chunk_size=DEFAULT_CHUNKSIZE):
        #     yield self._make_result(r, backend)

        yield self._query(txn, query_text, parameters, backend, kws)

    def _read_table_chunked(self, txn, table_name, backend, kws):
        query_text = self._select_all_from_table_query(table_name)
        return self._query_chunked(txn, query_text, None, backend, kws)




import pathlib

from plpipes.exceptions import AuthenticationError
from plpipes.plugin import Plugin

class AuthenticatorBase(Plugin):

    def __init__(self, account_name, acfg):
        self._account_name = account_name
        self._cfg = acfg
        self._credentials = None

    def credentials(self):
        if self._credentials is None:
            try:
                self._credentials = self._authenticate()
            except AuthenticationError:
                raise
            except Exception as ex:
                raise AuthenticationError(f"Authentication for Azure account {self._account_name} failed") from ex
        return self._credentials

    def _private_path(self, subdir=None, create=True):
        path = pathlib.Path.home() / f".config/plpipes/cloud/azure/auth" / self._account_name
        if subdir is not None:
            path = path / subdir
        if create:
            path.mkdir(exist_ok=True, parents=True)
        return path

    def _credentials_cache_filename(self):
        return self._private_path() / "cached.json"




"""
This module handles the registration and execution of various action drivers
for the plpipes framework.

Action drivers are responsible for executing specific types of actions defined
within the project. The supported action drivers include simple actions, SQL
actions, downloading actions, quarto processing, file downloading, archive
unpacking, and loop actions.

"""

from .driver import simple
from .driver import sql
from .driver import downloader
from .driver import quarto
from .driver import file_downloader
from .driver import archive_unpacker
from .driver import loop

# import the runner
from .runner import run



from plpipes.config import cfg
import plpipes.cloud.google.auth

c = cfg.cd("cloud.google.vertexai")
creds = plpipes.cloud.google.auth.credentials(c["auth"])

import vertexai
vertexai.init(project=c["project"], location=c["location"], credentials=creds)

import sys
sys.modules[__name__] = vertexai



import logging
import time
import re

class Action:
    """
    Represents an action to be performed within the plpipes framework.

    Attributes:
        _name (str): The name of the action.
        _cfg (dict): Configuration parameters for the action.
    """

    def __init__(self, name, action_cfg):
        """
        Initializes the Action instance.

        Args:
            name (str): The name of the action.
            action_cfg (dict): Configuration parameters for the action.
        """
        self._name = name
        self._cfg = action_cfg

    def name(self):
        """
        Returns the name of the action.

        Returns:
            str: The name of the action.
        """
        return self._name

    def short_name(self):
        """
        Returns the short name of the action, derived from the full name.

        Returns:
            str: The short name of the action.
        """
        return re.split(r'[./\\]', self._name)[-1]

    def _do_it(self, indent):
        """
        Executes the action.

        Args:
            indent (int): The indentation level for logging.
        """
        self.do_it()

    def do_it(self):
        """
        The main logic of the action must be implemented in subclasses.
        This should contain the code that performs the action.
        """
        ...

    def run(self, indent=0):
        """
        Executes the action and logs its execution time.

        Args:
            indent (int): The indentation level for logging.
        """
        name = self.name()
        logging.info(f"{' '*indent}Action {name} started")
        start = time.time()
        self._do_it(indent=indent)
        lapse = int(10 * (time.time() - start) + 0.5) / 10.0
        logging.info(f"{' '*indent}Action {name} done ({lapse}s)")

    def __str__(self):
        """
        Returns a string representation of the Action instance.

        Returns:
            str: A string representation of the Action.
        """
        return f"<Action {self._name}>"



import logging

_class_registry = {}

_suffix_registry = []  # vector of pairs (suffix, action type)

def register_class(action_type, action_class, *suffixes):
    """
    Register a new action class with its associated action type and suffixes.

    Parameters:
    action_type (str): The type of action being registered.
    action_class (type): The class that implements the action.
    suffixes (str): One or more file suffixes associated with the action type.
    """
    _class_registry[action_type] = action_class

    for suffix in suffixes:
        _suffix_registry.append((suffix, action_type))

    _suffix_registry.sort(key=lambda x: len(x[0]), reverse=True)

def _action_type_lookup(files):
    """
    Lookup the action type based on the provided file suffixes.

    Parameters:
    files (list): A list of file names to check against registered suffixes.

    Returns:
    str: The action type associated with the matched suffix, or None if no match is found.
    """
    logging.debug(f"suffix registry: {_suffix_registry}")

    for suffix, action_type in _suffix_registry:
        if suffix in files:
            return action_type
    return None

def _action_class_lookup(type):
    """
    Retrieve the action class associated with the given action type.

    Parameters:
    type (str): The action type to look up.

    Returns:
    type: The class associated with the action type.

    Raises:
    ValueError: If the action type is not supported.
    """
    if type in _class_registry:
        return _class_registry[type]
    raise ValueError(f"Unsupported action type {type}")




from ..helpers import download_xlsx

def download(path, params, db, acfg):
    if path == "codmun20":
        data = download_xlsx("https://www.ine.es/daco/daco42/codmun/codmun20/20codmun.xlsx", db, "ine_codmun20", header=1)

    else:
        raise ValueError("Unknown/unsupported path for INE downloader")



import logging

from plpipes.config import cfg
from plpipes.action.base import Action
from plpipes.action.registry import register_class
from plpipes.action.runner import lookup
from plpipes.init import init_run_as_of_date
import plpipes

class _Iterator:
    """
    Base class for iterators used in the plpipes framework.

    Attributes:
        _key (str): The key associated with the iterator.
        _cfg (dict): Configuration dictionary for the iterator.
    """

    def __init__(self, key, icfg):
        """
        Initializes the iterator with a key and configuration.

        Args:
            key (str): The key for the iterator.
            icfg (dict): Configuration settings for the iterator.
        """
        self._key = key
        self._cfg = icfg
        self.reset()

    def reset(self):
        """Resets the iterator to its initial state."""
        pass

    def next(self):
        """
        Advances the iterator to the next value.

        Returns:
            bool: True if there is a next value; False otherwise.
        """
        return False

    def where(self):
        """
        Provides a string representation of the current state of the iterator.

        Returns:
            str: A string indicating the current key and value.
        """
        return self._key

class _ListIterator(_Iterator):
    """
    Iterator for a list of values.

    Attributes:
        _values (list): The list of values to iterate over.
        _target (str): The target configuration key to set on each iteration.
    """

    def __init__(self, key, icfg, values):
        """
        Initializes the list iterator.

        Args:
            key (str): The key for the iterator.
            icfg (dict): Configuration settings for the iterator.
            values (list): The list of values to iterate over.
        """
        self._values = values
        self._target = icfg["target"]
        super().__init__(key, icfg)

    def reset(self):
        """Resets the iterator to the start of the values."""
        self._ix = -1
        cfg[self._target] = None

    def next(self):
        """
        Advances the iterator to the next value.

        Returns:
            bool: True if there is a next value; False otherwise.
        """
        self._ix += 1
        try:
            v = self._values[self._ix]
        except IndexError:
            return False
        logging.debug(f"Setting {self._target} to {v}")
        cfg[self._target] = v
        return True

    def where(self):
        """
        Provides a string representation of the current state of the iterator.

        Returns:
            str: A string indicating the current key and value from _values.
        """
        return f"{self._key}={self._values[self._ix]}"

class _ValuesIterator(_ListIterator):
    """
    Iterator for a list of values obtained from the configuration.

    Args:
        key (str): The key for the iterator.
        icfg (dict): Configuration settings for the iterator.
    """

    def __init__(self, key, icfg):
        """
        Initializes the values iterator.

        Args:
            key (str): The key for the iterator.
            icfg (dict): Configuration settings for the iterator.
        """
        super().__init__(key, icfg, list(icfg["values"]))

class _ConfigKeysIterator(_ListIterator):
    """
    Iterator for configuration keys.

    Args:
        key (str): The key for the iterator.
        icfg (dict): Configuration settings for the iterator.
    """

    def __init__(self, key, icfg):
        """
        Initializes the config keys iterator.

        Args:
            key (str): The key for the iterator.
            icfg (dict): Configuration settings for the iterator.
        """
        values = list(cfg.cd(icfg["path"]).keys())
        super().__init__(key, icfg, values)

class RunAsOfDateIterator(_ListIterator):
    """
    Iterator for run-as-of-date values based on specified parameters.

    Args:
        key (str): The key for the iterator.
        icfg (dict): Configuration settings for the iterator.
    """

    def __init__(self, key, icfg):
        """
        Initializes the run-as-of-date iterator.

        Args:
            key (str): The key for the iterator.
            icfg (dict): Configuration settings for the iterator.
        """
        values = icfg.get("values")
        icfg.setdefault("target", "run.as_of_date")

        start = icfg.get("start")
        if start is not None:
            end = icfg.get("end", "today")
            periodicity = icfg.get("periodicity", "daily")
            values = self._date_range(start, end, periodicity, values)
        elif values is None:
            raise ValueError("No values or start date provided for RunAsOfDateIterator")

        super().__init__(key, icfg, list(values))

    def _date_range(self, start, end, periodicity, more_values):
        """
        Generates a range of dates based on the specified parameters.

        Args:
            start (str): Start date.
            end (str): End date.
            periodicity (str): The frequency of the dates (daily, weekly, etc.).
            more_values (list): Additional values to include in the range.

        Returns:
            list: A sorted list of dates within the specified range.
        """
        from friendlydateparser import parse_date
        from dateutil.relativedelta import relativedelta
        start = parse_date(start)
        end = parse_date(end)
        if periodicity == "daily":
            step = relativedelta(days=1)
        elif periodicity == "weekly":
            step = relativedelta(days=7)
        elif periodicity == "monthly":
            step = relativedelta(months=1)
        elif periodicity == "yearly":
            step = relativedelta(years=1)
        else:
            raise ValueError(f"Unsupported periodicity {periodicity}")

        if more_values is None:
            more_values = []
        values = set(fdp.parse_date(v) for v in more_values)
        value = start
        while value <= end:
            values.add(value)
            value += step
        values = sorted(values)
        logging.debug(f"Date range from {start} to {end} with periodicity {periodicity} yields {values}")
        return values

    def reset(self):
        """Resets the current date to 'now' and initializes the run date."""
        super().reset()
        cfg['run.as_of_date'] = 'now'
        init_run_as_of_date()

    def next(self):
        """
        Advances the iterator to the next date value.

        Returns:
            bool: True if there is a next date; False otherwise.
        """
        if super().next():
            init_run_as_of_date()
            return True
        return False

def _init_iterator(key, icfg):
    """
    Initializes an iterator based on the specified type in configuration.

    Args:
        key (str): The key for the iterator.
        icfg (dict): Configuration settings for the iterator.

    Returns:
        _Iterator: An instance of a specific iterator class.

    Raises:
        NotImplementedError: If the iterator type is unsupported.
    """
    type = icfg.get("type", "value")
    if type == "values":
        return _ValuesIterator(key, icfg)
    elif type == "configkeys":
        return _ConfigKeysIterator(key, icfg)
    elif type == "runasofdate":
        return RunAsOfDateIterator(key, icfg)
    else:
        raise NotImplementedError(f"Unsupported iterator type {type} found in loop")

def _iterate(iterators):
    """
    Iterates through a list of iterators and yields the state of each.

    Args:
        iterators (list): List of iterator instances to be iterated.

    Yields:
        str: A string representation of the current state of all iterators.
    """
    level = 0
    while level >= 0:
        if iterators[level].next():
            if level < len(iterators) - 1:
                level += 1
            else:
                wheres = [i.where() for i in iterators]
                yield "/".join(wheres)
        else:
            iterators[level].reset()
            level -= 1

class _Loop(Action):
    """
    Action class that represents a loop of operations, utilizing specified iterators.

    Methods:
        do_it: Executes the loop action based on the configured sequence of operations.
    """

    def do_it(self):
        """
        Executes the loop action, iterating through the configured children actions.
        Logs iteration states and handles any exceptions based on configuration.
        """
        name = self._name

        children = [lookup(name, parent=self._name)
                    for name in self._cfg["sequence"]]

        iterators = []
        iicfg = self._cfg.cd("iterator")
        for key in iicfg.keys():
            icfg = iicfg.cd(key)
            iterators.append(_init_iterator(key, icfg))

        for where in _iterate(iterators):
            logging.info(f"Iterating at {where}")
            try:
                for child in children:
                    child.run()
            except Exception as ex:
                if self._cfg.get("ignore_errors", False):
                    logging.exception(f"Iteration {where} failed")
                else:
                    raise

register_class("loop", _Loop)



from plpipes.config import cfg
import plpipes.plugin

_backend_class_registry = plpipes.plugin.Registry("google_auth_backend",
                                                  "plpipes.cloud.google.auth.plugin")

_registry = {}

def credentials(account_name):
    if account_name not in _registry:
        _registry[account_name] = _init_backend(account_name)
    return _registry[account_name].credentials()

def _init_backend(account_name):
    acfg = cfg.cd("cloud.google.auth").cd(account_name)

    backend_name = acfg.get("driver", "oauth2")
    backend_class = _backend_class_registry.lookup(backend_name)
    return backend_class(account_name, acfg)




def download_json(url, headers={}):
    return requests.get(url, headers=headers).json()

def download_to_file(url):
    tmp = tempfile.NamedTemporaryFile(delete=False)
    r = requests.get(url)
    tmp.write(r.content)
    tmp.close()
    return tmp.name




"""
Module: plpipes.runner

The purpose of the runner is to offer a unified entry point for the
project actions and pipelines. It extracts information from a set of
environment variables and parses command line arguments in a standard way.

For more information, refer to [Runner](../../runner.md).
"""

import plpipes.init
import plpipes.action
import plpipes.config
import sys
import argparse
import os
import pathlib
import json
import logging

class _PairAction(argparse.Action):
    """
    Custom argparse action for handling key-value pairs input.

    This action allows key-value pairs to be passed in as arguments,
    with optional unpacking for JSON values.
    """
    def __init__(self, option_strings, dest, nargs=None, unpack=None, default=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        if default is None:
            default = {}
        super().__init__(option_strings, dest, nargs=1, default=default, **kwargs)
        self.unpack = unpack

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Processes the input key-value pair(s) and assigns them to the namespace.

        Parameters:
            parser: The argument parser.
            namespace: The namespace where the values should be stored.
            values: The input values to process.
            option_string: The option string used for this action.
        """
        if self.unpack:
            if self.unpack == "json":
                def unpacker(x):
                    return json.loads(x)
            else:
                raise Exception(f"Bad unpacker {self.unpack}")
        else:
            unpacker = None
        for pair in values:
            try:
                (k, v) = pair.split("=", 2)
            except:
                raise argparse.ArgumentError(self, f"Can not parse config pair '{pair}'")
            if unpacker:
                try:
                    v = unpacker(v)
                except:
                    raise argparse.ArgumentError(self, f"Can not unpack value part of '{pair}' as {self.unpack}.")
            try:
                getattr(namespace, self.dest).append({k: v})
            except Exception as ex:
                raise argparse.ArgumentError(self, f"Conflicting config pair '{pair}': {ex}") from ex


_default_arg_parser_args = { 'description': 'PLPipes runner' }

def arg_parser(**kwargs):
    """
    Initializes the argument parser for the runner.

    Parameters:
        kwargs: Additional keyword arguments for customizing the parser.

    Returns:
        An instance of argparse.ArgumentParser configured for the runner.
    """
    parser = argparse.ArgumentParser(**{**_default_arg_parser_args, **kwargs})
    parser.add_argument('-d', '--debug',
                        help="Turns on debugging",
                        action='store_true')
    parser.add_argument('-c', '--config',
                        action="append",
                        metavar="CFG_FN",
                        help="Additional configuration file",
                        default=[])
    parser.add_argument('-s', '--set',
                        action=_PairAction,
                        metavar="CFG_KEY=VAL",
                        help="Set configuration entry",
                        default=[])
    parser.add_argument('-S', '--set-json',
                        action=_PairAction,
                        metavar="CFG_KEY=JSON_VAL",
                        unpack="json",
                        dest="set",
                        help="Set configuration entry (value is parsed as JSON)")
    parser.add_argument('-e', '--env',
                        metavar="ENVIRONMENT",
                        help="Select environment (dev, pre, pro, etc.)")
    return parser

def parse_args_and_init(arg_parser, args=None):
    """
    Parses command line arguments and initializes the PLPipes framework.

    Parameters:
        arg_parser: The argument parser instance to use for parsing.
        args: The command line arguments to parse (defaults to sys.argv).

    Returns:
        The parsed options as an object with attributes corresponding to the arguments.

    Raises:
        Exception: If the program name is missing from the argument list.
    """
    if args is None:
        args = sys.argv

    if len(args) < 1:
        raise Exception("Unable to infer config stem. Program name missing from argument list")

    prog_path = pathlib.Path(args[0])
    if "PLPIPES_ROOT_DIR" in os.environ:
        root_dir = pathlib.Path(os.environ["PLPIPES_ROOT_DIR"]).resolve(strict=True)
    else:
        root_dir = prog_path.parent.parent
    root_dir = root_dir.absolute()

    opts = arg_parser.parse_args(args[1:])

    config_extra = [{'fs': {'stem': str(prog_path.stem),
                            'root': str(root_dir),
                            'project': str(root_dir.stem)}}]

    config_extra += opts.set

    if opts.env is not None:
        config_extra["env"] = opts.env
    if opts.debug:
        config_extra.append({"logging.level": "debug"})
        config_extra.append({"logging.level_file": "debug"})

    plpipes.init.init(*config_extra, config_files=opts.config)

    sys.path.append(plpipes.config.cfg["fs.lib"])

    os.environ.setdefault("PLPIPES_ROOT_DIR", plpipes.config.cfg["fs.root"])

    return opts

def simple_init(args=None):
    """
    A simplified initialization function that uses a default argument parser.

    Parameters:
        args: The command line arguments to parse (defaults to None).

    Returns:
        The parsed options as an object with attributes corresponding to the arguments.
    """
    return parse_args_and_init(arg_parser(), args)

def main(args=None):
    """
    Main entry point for the runner. Parses arguments and executes specified actions.

    Parameters:
        args: The command line arguments to parse (defaults to None).
    """
    parser = arg_parser()
    parser.add_argument('actions', nargs="*",
                        metavar="ACTION", default=["default"])
    opts = parse_args_and_init(parser, args)

    for action in opts.actions:
        logging.info(f"Executing action {action}")
        plpipes.action.run(action)



import logging
import geopandas
import pandas

from plpipes.database.backend.pandas import PandasBackend
from plpipes.util.database import split_table_name

DEFAULT_CHUNKSIZE = 5000

class GeoPandasBackend(PandasBackend):

    def register_handlers(self, handlers):
        super().register_handlers(handlers)
        handlers["create_table"].register(geopandas.GeoDataFrame, self._create_table_from_geopandas)

    def _create_table_from_geopandas(self, txn, table_name, df, parameters, if_exists, kws):
        logging.debug("Creating table from geopandas")
        chunksize = txn._driver._pop_kw(kws, "chunksize", DEFAULT_CHUNKSIZE)
        schema, table_name = split_table_name(table_name)
        return df.to_postgis(table_name, txn._conn,
                             schema="main" if schema is None else schema,
                             if_exists=if_exists,
                             index=False, chunksize=chunksize, **kws)

    def _df_read_sql(self, txn, sqla, geom_col=None, **kws):
        if geom_col is None:
            return pandas.read_sql(sqla, txn._conn, **kws)
        else:
            return geopandas.read_postgis(sqla, txn._conn, geom_col=geom_col, **kws)

    def _df_concat(self, dfs):
        if isinstance(dfs[0], geopandas.GeoDataFrame):
            return geopandas.concat(dfs)
        else:
            return pandas.concat(dfs)



import logging
from plpipes.database.driver.odbc import ODBCDriver
import struct
import sqlalchemy.sql as sas

_SQL_COPT_SS_ACCESS_TOKEN = 1256

class AzureSQLDriver(ODBCDriver):
    def __init__(self, name, drv_cfg):
        driver = drv_cfg.setdefault('driver_odbc', '{ODBC Driver 18 for SQL Server}')
        server = drv_cfg['server']
        database = drv_cfg['database']
        low_level_proto = drv_cfg.setdefault('low_level_proto', 'tcp')
        port = drv_cfg.setdefault('port', 1433)
        cs = { 'Driver': driver,
               'Server': server,
               'Database': database }
        # cs = f"Driver={driver};Server={server};Database={database}"
        connect_args = {}

        if (((pwd := drv_cfg.get('password')) is not None) and
            ((uid := drv_cfg.getany('uid', 'user')) is not None)):
            cs['Pwd'] = pwd
            cs['UID'] = uid

        elif (cred_account_name := drv_cfg.get('credentials')) is not None:
            import plpipes.cloud.azure.auth
            credential = plpipes.cloud.azure.auth.credentials(creds_account_name)

            token = credential.get_token('https://database.windows.net').token.encode("utf-16-le")
            token_struct = struct.pack(f"<I{len(token)}s", len(token), token)

        else:
            raise ValueError(f"Authentication data missing from configuration for database instance {name}")


        cs = ";".join(f"{k}={v}" for k, v in cs.items())
        drv_cfg.setdefault('connection_string', cs)
        drv_cfg.setdefault('sql_alchemy_driver', 'mssql+pyodbc')

        super().__init__(name, drv_cfg, connect_args=connect_args)

    def _list_tables_query(self):
        table = sas.table("TABLES", schema="INFORMATION_SCHEMA")
        return (sas.select(sas.column("table_name").label("name"))
                .select_from(table)
                .where(sas.column("table_type") == "BASE TABLE"))

    def _list_views_query(self):
        table = sas.table("TABLES", schema="INFORMATION_SCHEMA")
        return (sas.select(sas.column("table_name").label("name"))
                .select_from(table)
                .where(sas.column("table_type") == "VIEW"))



import geopandas
import pandas
import random
import sqlalchemy.sql as sas
from sqlalchemy.sql.expression import func as saf

from plpipes.database.backend.geopandas import GeoPandasBackend
from plpipes.database.sqlext import AsSubquery
from plpipes.plugin import plugin

@plugin
class GeoPandasSpatialiteBackend(GeoPandasBackend):

    def _df_read_sql(self, txn, sqla, wkb_geom_col=None, geom_col=None, **kws):
        if geom_col is not None:
            wrapped_col = f"{geom_col}__wrapped{random.randint(0,10000)}"
            wrapped_sql = (sas.select("*",
                                      saf.Hex(saf.ST_AsBinary(sas.column(geom_col))).label(wrapped_col))
                           .select_from(AsSubquery(sqla)))

            df = geopandas.read_postgis(wrapped_sql, txn._conn, geom_col=wrapped_col, **kws)
            df.drop([geom_col], axis=1, inplace=True)
            df.rename(columns={wrapped_col: geom_col}, inplace=True)
            df.set_geometry(geom_col, inplace=True)
            return df
        elif wkb_geom_col is not None:
            return geopandas.read_postgis(wrapped_sql, txn._conn, geom_col=wkb_geom_col, **kws)
        else:
            return pandas.read_sql(sqla, txn._conn, **kws)



import sqlalchemy.engine

from plpipes.database.driver.sqlalchemy import SQLAlchemyDriver

class ODBCDriver(SQLAlchemyDriver):
    def __init__(self, name, drv_cfg, **kwargs):
        url = sqlalchemy.engine.URL.create(drv_cfg['sql_alchemy_driver'],
                                           query={'odbc_connect': drv_cfg['connection_string']})
        super().__init__(name, drv_cfg, url, **kwargs)



import eurostat

import pandas as pd

def eu_hicp():
    df = eurostat.get_data_df("PRC_HICP_MIDX")
    df = df.melt(id_vars=['freq', 'unit', 'coicop', 'geo\TIME_PERIOD'],
                 var_name='period',
                 value_name='value')
    df.rename(columns={"geo\TIME_PERIOD": "geo__time_period"}, inplace=True)
    df["period"] = pd.to_datetime(df["period"]).dt.date
    return df




class Transaction:
    """
    The Transaction class represents a database transaction.

    Attributes:
        driver (object): The database driver object.
        conn (object): The database connection object.
    """

    def __init__(self, driver, conn):
        """
        Creates a new transaction object.

        Args:
            driver (object): The database driver object.
            conn (object): The database connection object.

        Note:
            Transaction objects should not be created calling the class
            constructor directly but through Driver `begin` method.
        """
        self._driver = driver
        self._conn = conn

    def driver(self):
        """
        Returns the database driver object associated with this transaction.

        Returns:
            object: The database driver object.
        """
        return self._driver

    def db_name(self):
        """Returns the name of the database."""
        return self._driver._name

    def connection(self):
        """
        Returns the database connection object associated with this transaction.

        Returns:
            object: The database connection object.
        """
        return self._conn

    def execute(self, sql, parameters=None):
        """
        Executes an SQL statement with optional parameters.

        Args:
            sql (str): The SQL statement to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
        """
        self._driver._execute(self, sql, parameters)

    def execute_script(self, sql_script):
        """
        Executes a script containing multiple SQL statements.

        Args:
            sql_script (str): The SQL script to execute.
        """
        return self._driver._execute_script(self, sql_script)

    def create_table(self, table_name, sql_or_df, parameters=None, if_exists="replace", **kws):
        """
        Creates a new table in the database.

        Args:
            table_name (str): The name of the table to create.
            sql_or_df (str or DataFrame): The SQL statement or DataFrame defining the table schema.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            if_exists (str, optional): How to handle the table if it already exists. Valid options are "fail", "replace", and "append".
            **kws: Additional keyword arguments to pass to the driver.
        """
        return self._driver._create_table(self, table_name, sql_or_df, parameters, if_exists, kws)

    def create_view(self, view_name, sql, parameters=None, if_exists="replace", **kws):
        """
        Creates a new view in the database.

        Args:
            view_name (str): The name of the view to create.
            sql (str): The SQL statement defining the view.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            if_exists (str, optional): How to handle the view if it already exists. Valid options are "fail", "replace", and "append".
            **kws: Additional keyword arguments to pass to the driver.
        """
        return self._driver._create_view(self, view_name, sql, parameters, if_exists, kws)

    def read_table(self, table_name, backend=None, **kws):
        """
        Reads a table from the database into a DataFrame.

        Args:
            table_name (str): The name of the table to read.
            backend (optional): The backend to use for reading the table. If None, the default backend for the driver is used.
            **kws: Additional keyword arguments to pass to the backend.

        Returns:
            DataFrame: A DataFrame containing the table data.
        """
        return self._driver._read_table(self, table_name, backend, kws)

    def read_table_chunked(self, table_name, backend=None, **kws):
        """
        Reads a table from the database in chunks.

        Args:
            table_name (str): The name of the table to read.
            backend (optional): The backend to use for reading the table. If None, the default backend for the driver is used.
            **kws: Additional keyword arguments to pass to the backend.
        """
        return self._driver._read_table_chunked(self, table_name, backend, kws)

    def query(self, sql, parameters=None, backend=None, **kws):
        """
        Executes an SQL query and returns the result as a DataFrame.

        Args:
            sql (str): The SQL query to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            backend (optional): The backend to use for executing the query. If None, the default backend is used.
            **kws: Additional keyword arguments to pass to the driver.

        Returns:
            DataFrame: A DataFrame containing the query result.
        """
        return self._driver._query(self, sql, parameters, backend, kws)

    def query_first(self, sql, parameters=None, backend=None, **kws):
        """
        Executes an SQL query and returns the first row of the result.

        Args:
            sql (str): The SQL query to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            backend (optional): The backend to use for executing the query. If None, the default backend is used.
            **kws: Additional keyword arguments to pass to the driver.

        Returns:
            DataFrame or dict: A dataframe/dictionary containing the result first row.
        """
        return self._driver._query_first(self, sql, parameters, backend, kws)

    def query_first_value(self, sql, parameters=None, backend="tuple", **kws):
        """
        Executes an SQL query and returns the first value from the result.

        Args:
            sql (str): The SQL query to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            backend (str, optional): The backend to use for executing the query. If None, the default backend is used. Defaults to `tuple`.
            **kws: Additional keyword arguments to pass to the driver.

        Returns:
            Any: The first value from the result (first row, first column).
        """
        return self._driver._query_first_value(self, sql, parameters, backend, kws)

    def query_chunked(self, sql, parameters=None, backend=None, **kws):
        """
        Executes an SQL query and returns the result as an iterator over chunks of rows.

        Args:
            sql (str): The SQL query to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            backend (optional): The backend to use for executing the query. If None, the default backend is used.
            **kws: Additional keyword arguments to pass to the driver.

        Returns:
            iterator: An iterator over chunks of rows.
        """
        return self._driver._query_chunked(self, sql, parameters, backend, kws)

    def query_group(self, sql, parameters=None, by=None, backend=None, **kws):
        """
        Executes an SQL query and returns the result as a DataFrame grouped by one or more columns.

        Args:
            sql (str): The SQL query to execute.
            parameters (dict, optional): A dictionary containing values to fill in SQL statement placeholders.
            by (optional): The column(s) to group by.
            backend (optional): The backend to use for executing the query. If None, the default backend is used.
            **kws: Additional keyword arguments to pass to the driver.

        Returns:
            DataFrame: A DataFrame containing the grouped query result.
        """
        return self._driver._query_group(self, sql, parameters, by, backend, kws)

    def drop_table(self, table_name, only_if_exists=True):
        """
        Drops a table from the database.

        Args:
            table_name (str): The name of the table to drop.
            only_if_exists (bool, optional): If True, the table is only dropped if it exists. Otherwise, an error is raised if the table does not exist.
        """
        return self._driver._drop_table(self, table_name, only_if_exists)

    def list_tables(self):
        """
        Lists the tables in the database.

        Returns:
            DataFrame: DataFrame with the list of tables.
        """
        return self._driver._list_tables(self)

    def list_views(self):
        """
        Lists the views in the database.

        Returns:
            DataFrame: DataFrame with the list of views.
        """
        return self._driver._list_views(self)

    def table_exists_p(self, table_name):
        """
        Checks whether a table exists in the database.

        Args:
            table_name (str): The name of the table to check.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        return self._driver._table_exists_p(self, table_name)

    def copy_table(self, from_table_name, to_table_name, if_exists="replace", **kws):
        """
        Copies the contents of one table to another.

        Args:
            from_table_name (str): The name of the table to copy from.
            to_table_name (str): The name of the table to copy to.
            if_exists (str, optional): How to handle the destination table if it already exists. Valid options are "fail", "replace", and "append".
            **kws: Additional keyword arguments to pass to the driver.

        Raises:
            ValueError: If the source and destination table names are the same.

        Returns:
            int: The number of rows copied.
        """
        if from_table_name == to_table_name:
            raise ValueError("source and destination tables must be different")
        return self._driver._copy_table(self, from_table_name, to_table_name, if_exists, kws)



from plpipes.plugin import plugin
from plpipes.database.backend.spark import SparkBackendBase
from plpipes.spark import spark_session

import pandas

@plugin
class PandasSparkHiveBatckend(SparkBackendBase):
    def _coerce_output_df(self, df):
        return df.toPandas()

    def _create_table_from_pandas(self, txn, table_name, df, *args):
        df = spark_session().createDataFrame(df)
        return self._create_table_from_spark(txn, table_name, df, *args)

    def register_handlers(self, handlers):
        handlers['create_table'].register(pandas.DataFrame, self._create_table_from_pandas)



from plpipes.plugin import Plugin, plugin

from databricks.connect import DatabricksSession

@plugin
class DatabricksPlugin(Plugin):

    def init_spark_session(self, ssc):
        profile = ssc.get("profile", "DEFAULT")
        builder = DatabricksSession.builder.profile(profile)

        for k, v in ssc.to_flat_dict('extra').items():
            builder = builder.config(k, v)

        ss = builder.getOrCreate()

        default_database = ssc.get("default_database", None)
        if default_database is not None:
            ss.sql(f"USE {default_database}")

        return ss



from plpipes.plugin import plugin
from plpipes.tool.dbeaver.conarg import ConArg

class SQLServerConArg(ConArg):
    def __init__(self, name, db_drv):
        super().__init__(name, db_drv)
        db_cfg = self._cfg
        host = db_cfg['server']
        database = db_cfg['database']
        port = db_cfg.get('port', '1433')
        user = db_cfg.get('uid')
        password = db_cfg.get('password')

        self.host = host
        self.server = host
        self.port = port
        self.database = database
        self.driver = "microsoft"
        self.auth = "native"
        self.user = user
        self.password = password



"""
Module for handling database drivers in the plpipes framework.

This module defines the Driver class that acts as an abstraction layer for
database interactions, allowing for the use of various backends for
database operations. It provides an interface for executing SQL commands,
managing transactions, and interacting with tables and views within a
database.

Classes:
    Driver: The base class for database drivers providing methods for
            executing SQL commands, handling transactions, and managing
            database backend interactions.
"""

import logging
from contextlib import contextmanager
from plpipes.util.typedict import dispatcher
from plpipes.util.method_decorators import optional_abstract
import plpipes.plugin
import types
import plpipes.database.driver.transaction

_backend_class_registry = plpipes.plugin.Registry("db_backend", "plpipes.database.backend.plugin")

class Driver(plpipes.plugin.Plugin):
    """
    Driver class for handling database interactions.

    Attributes:
        _default_backend_name (str): The default backend name to use.
        _backend_subkeys (list): List of backend subkeys associated with this driver.

    Methods:
        config(): Returns the configuration for the driver.
        driver_name(): Returns the name of the driver.
        begin(): Context manager for starting a transaction.
        _execute(txn, sql, parameters=None): Executes an SQL command in the transaction.
        _execute_script(txn, sql): Executes a SQL script in the transaction.
        _list_tables(txn): Lists tables in the database.
        _read_table(txn, table_name, backend, kws): Reads a table from the database.
        _drop_table(txn, table_name, only_if_exists): Drops a table from the database.
        _create_table(txn, table_name, sql_or_df, parameters, if_exists, kws): Creates a new table.
        _create_view(txn, view_name, sql, parameters, if_exists, kws): Creates a new view in the database.
        _copy_table(txn, from_table_name, to_table_name, if_exists, kws): Copies data between tables.
        _query_chunked(txn, sql, parameters, backend, kws): Executes a chunked query.
        _query_group(txn, sql, parameters, by, backend, kws): Executes a grouped query.
        load_backend(name): Loads a specific backend into the driver.
    """

    _default_backend_name = "pandas"
    _backend_subkeys = []

    @classmethod
    def _init_plugin(klass, key):
        """
        Initializes the plugin with backend registry and configuration.

        Args:
            klass: The class reference.
            key: The key of the plugin instance.
        """
        super()._init_plugin(key)
        klass._backend_registry = {}
        klass._backend_subkeys = [key, *klass._backend_subkeys]
        klass._create_table = klass._create_table.copy()

    @classmethod
    def _backend_lookup(klass, name):
        """
        Looks up and returns the specified backend by name.

        Args:
            klass: The class reference.
            name: The name of the backend to look up.

        Returns:
            backend: The backend instance associated with the specified name.
        """
        try:
            return klass._backend_registry[name]
        except KeyError:
            backend_class = _backend_class_registry.lookup(name, subkeys=klass._backend_subkeys)
            backend = backend_class()
            klass._backend_registry[name] = backend
            backend.register_handlers({'create_table': klass._create_table.td})
            logging.debug(f"backend {backend._plugin_name} for {klass._plugin_name} loaded")
            return backend

    def __init__(self, name, drv_cfg):
        """
        Initializes the Driver instance with a name and configuration.

        Args:
            name: The name of the database driver.
            drv_cfg: The configuration settings for the driver instance.
        """
        self._name = name
        self._cfg = drv_cfg
        self._last_key = 0
        self._default_backend = self._backend_lookup(self._cfg.get("backend", self._default_backend_name))
        for backend_name in self._cfg.get('extra_backends', []):
            self._backend_lookup(backend_name)

    def config(self):
        """
        Returns the configuration settings for the driver.

        Returns:
            Tree: A tree representation of the configuration.
        """
        return self._cfg.to_tree()

    def _backend(self, name):
        """
        Retrieves the requested backend or the default backend if none is specified.

        Args:
            name: The name of the backend to retrieve.

        Returns:
            backend: The specified backend instance or default backend if name is None.
        """
        if name is None:
            return self._default_backend
        logging.debug(f"looking up backend {name}")
        return self._backend_lookup(name)

    def driver_name(self):
        """
        Returns the name of the database driver.

        Returns:
            str: The name of the driver.
        """
        return self._plugin_name

    @optional_abstract
    @contextmanager
    def begin(self):
        """
        Context manager for beginning a database transaction.

        Yields:
            Transaction: A transaction object for conducting operations within a context.
        """
        ...

    @optional_abstract
    def _execute(self, txn, sql, parameters=None):
        """
        Executes an SQL command within a transaction.

        Args:
            txn: The transaction instance to execute the command within.
            sql: The SQL command to execute.
            parameters: Optional parameters for the SQL command.
        """
        ...

    @optional_abstract
    def _execute_script(self, txn, sql):
        """
        Executes a SQL script within a transaction.

        Args:
            txn: The transaction instance to execute the script within.
            sql: The SQL script to execute.
        """
        ...

    @optional_abstract
    def _list_tables(self, txn):
        """
        Lists all tables within the database for the given transaction.

        Args:
            txn: The transaction instance for querying the database.
        """
        ...

    def _next_key(self):
        """
        Generates the next unique key for database operations.

        Returns:
            int: A unique key for the current operation.
        """
        self._last_key += 1
        return self._last_key

    def _query(self, txn, sql, parameters, backend, kws):
        """
        Executes a database query and returns the result.

        Args:
            txn: The transaction instance to use for the query.
            sql: The SQL query string to execute.
            parameters: Optional parameters for the SQL query.
            backend: The backend to use for executing the query.
            kws: Additional keyword arguments for the backend.

        Returns:
            Result: The result of the query execution.
        """
        logging.debug(f"database query code: {repr(sql)}, parameters: {str(parameters)[0:40]}")
        return self._backend(backend).query(txn, sql, parameters, kws)

    def _query_first(self, txn, sql, parameters, backend, kws):
        """
        Executes a query and returns the first result.

        Args:
            txn: The transaction instance.
            sql: The SQL query string to execute.
            parameters: Optional parameters for the SQL query.
            backend: The backend to use for executing the query.
            kws: Additional keyword arguments.

        Returns:
            Result: The first result of the query execution.
        """
        logging.debug(f"database query code: {repr(sql)}, parameters: {str(parameters)[0:40]}")
        return self._backend(backend).query_first(txn, sql, parameters, kws)

    def _query_first_value(self, txn, sql, parameters, backend, kws):
        """
        Executes a query and returns the first value from the result.

        Args:
            txn: The transaction instance.
            sql: The SQL query string to execute.
            parameters: Optional parameters for the SQL query.
            backend: The backend to use for executing the query.
            kws: Additional keyword arguments.

        Returns:
            Any: The first value from the result of the query execution.
        """
        logging.debug(f"database query code: {repr(sql)}, parameters: {str(parameters)[0:40]}")
        return self._backend(backend).query_first_value(txn, sql, parameters, kws)

    @optional_abstract
    def _read_table(self, txn, table_name, backend, kws):
        """
        Reads a table from the database.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to read.
            backend: The backend to use for reading the table.
            kws: Additional keyword arguments.

        Returns:
            DataFrame: The data read from the table.
        """
        ...

    @optional_abstract
    def _drop_table(self, txn, table_name, only_if_exists):
        """
        Drops a specified table from the database.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to be dropped.
            only_if_exists: Boolean to specify if the table should only be dropped if it exists.
        """
        ...

    @dispatcher({str: '_create_table_from_str',
                 list: '_create_table_from_records',
                 types.GeneratorType: '_create_table_from_iterator',},
                ix=2)
    def _create_table(self, txn, table_name, sql_or_df, parameters, if_exists, kws):
        """
        Creates a table in the database from various input types.

        Args:
            txn: The transaction instance to create the table within.
            table_name: The name of the table to create.
            sql_or_df: The SQL command or DataFrame to define the table.
            parameters: Optional parameters for creating the table.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        ...

    @optional_abstract
    def _create_table_from_str(self, txn, table_name, sql, parameters, if_exists, kws):
        """
        Creates a table from a SQL string command.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to create.
            sql: SQL string defining the table schema.
            parameters: Optional parameters for the SQL command.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        ...

    @optional_abstract
    def _create_table_from_clause(self, txn, table_name, clause, parameters, if_exists, kws):
        """
        Creates a table from a SQL clause.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to create.
            clause: SQL clause for creating the table.
            parameters: Optional parameters for the SQL command.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        ...

    def _create_table_from_records(self, txn, table_name, records, parameters, if_exists, kws):
        """
        Creates a table from a list of records.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to create.
            records: Iterable containing records for the table.
            parameters: Optional parameters for creating the table.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        backend = self._backend(kws.pop("backend", None))
        backend.create_table_from_records(txn, table_name, records, parameters, if_exists, kws)

    def _create_table_from_iterator(self, txn, table_name, iterator, parameters, if_exists, kws):
        """
        Creates a table from an iterator yielding records.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to create.
            iterator: Iterator yielding records for the table.
            parameters: Optional parameters for creating the table.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        for chunk in iterator:
            self._create_table(txn, table_name, chunk, parameters, if_exists, kws)
            if_exists = 'append'

    @optional_abstract
    def _create_view(self, txn, view_name, sql, parameters, if_exists, kws):
        """
        Creates a view in the database.

        Args:
            txn: The transaction instance.
            view_name: The name of the view to create.
            sql: SQL string defining the view.
            parameters: Optional parameters for creating the view.
            if_exists: Specifies how to handle the view if it already exists.
            kws: Additional keyword arguments.
        """
        ...

    @optional_abstract
    def _copy_table(self, txn, from_table_name, to_table_name, if_exists, kws):
        """
        Copies the contents of one table to another.

        Args:
            txn: The transaction instance.
            from_table_name: The name of the table to copy from.
            to_table_name: The name of the table to copy to.
            if_exists: Specifies how to handle the table if it already exists.
            kws: Additional keyword arguments.
        """
        ...

    @optional_abstract
    def _read_table_chunked(self, txn, table_name, backend, kws):
        """
        Reads a table in chunks.

        Args:
            txn: The transaction instance.
            table_name: The name of the table to read.
            backend: The backend to use for reading the table.
            kws: Additional keyword arguments.
        """
        ...

    def _query_chunked(self, txn, sql, parameters, backend, kws):
        """
        Executes a chunked query and returns results.

        Args:
            txn: The transaction instance.
            sql: The SQL query string to execute.
            parameters: Optional parameters for the SQL query.
            backend: The backend to use for executing the query.
            kws: Additional keyword arguments.

        Returns:
            Chunked results of the query execution.
        """
        return self._backend(backend).query_chunked(txn, sql, parameters, kws)

    def _query_group(self, txn, sql, parameters, by, backend, kws):
        """
        Executes a grouped query and returns results.

        Args:
            txn: The transaction instance.
            sql: The SQL query string to execute.
            parameters: Optional parameters for the SQL query.
            by: Column(s) to group the results by.
            backend: The backend to use for executing the query.
            kws: Additional keyword arguments.

        Returns:
            Grouped results of the query execution.
        """
        return self._backend(backend).query_group(txn, sql, parameters, by, kws)

    def load_backend(self, name):
        """
        Loads a specific backend into the driver.

        Args:
            name: The name of the backend to load.
        """
        self._backend_lookup(name)



"""
Module for defining the database backend interface in the plpipes framework.

This module provides the Backend class, which serves as an abstraction layer
for different database backends. It allows interaction with various data
representations such as pandas, geopandas, Spark DataFrames, polars, and more.

Classes:
    Backend: The base class for database backends providing methods for
             executing queries, handling data chunking, grouping results,
             and retrieving first results.
"""

from plpipes.plugin import Plugin

class Backend(Plugin):
    """
    Abstract base class for database backends.

    This class defines the interface for various database backends that can
    perform queries and manage data representations.

    Methods:
        query(engine, sql, parameters, kws): Executes a query and returns the result.
        query_chunked(engine, sql, parameter, kws): Executes a chunked query and returns results.
        query_group(engine, sql, parameters, by, kws): Executes a grouped query and returns results.
        query_first(engine, sql, parameters, kws): Executes a query and returns the first result.
        query_first_value(engine, sql, parameters, kws): Executes a query and returns the first value.
    """

    def query(self, engine, sql, parameters, kws):
        """
        Executes a query against the database.

        Args:
            engine (object): The database engine to use for executing the query.
            sql (str): The SQL query to execute.
            parameters (dict): Optional parameters for the SQL query.
            kws (dict): Additional keyword arguments for the execution.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError("This function is not yet implemented.")

    def query_chunked(self, engine, sql, parameters, kws):
        """
        Executes a chunked query against the database.

        Args:
            engine (object): The database engine to use for executing the query.
            sql (str): The SQL query to execute.
            parameters (dict): Optional parameters for the SQL query.
            kws (dict): Additional keyword arguments for the execution.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError("This function is not yet implemented.")

    def query_group(self, engine, sql, parameters, by, kws):
        """
        Executes a grouped query against the database.

        Args:
            engine (object): The database engine to use for executing the query.
            sql (str): The SQL query to execute.
            parameters (dict): Optional parameters for the SQL query.
            by (str): The column(s) to group the results by.
            kws (dict): Additional keyword arguments for the execution.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError("This function is not yet implemented.")

    def query_first(self, engine, sql, parameters, kws):
        """
        Executes a query and retrieves the first result.

        Args:
            engine (object): The database engine to use for executing the query.
            sql (str): The SQL query to execute.
            parameters (dict): Optional parameters for the SQL query.
            kws (dict): Additional keyword arguments for the execution.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError("This function is not yet implemented.")

    def query_first_value(self, engine, sql, parameters, kws):
        """
        Executes a query and retrieves the first value from the result.

        Args:
            engine (object): The database engine to use for executing the query.
            sql (str): The SQL query to execute.
            parameters (dict): Optional parameters for the SQL query.
            kws (dict): Additional keyword arguments for the execution.

        Raises:
            NotImplementedError: This method must be implemented in subclasses.
        """
        raise NotImplementedError("This function is not yet implemented.")



import pathlib
import logging
import re

from plpipes.config import cfg
from plpipes.action.registry import _action_class_lookup, _action_type_lookup

_action_cache = {}

def _find_action_files(action_root, name):
    """
    Find action files in the specified directory for a given action name.

    Args:
        action_root (pathlib.Path): The root directory to search for action files.
        name (str): The name of the action.

    Returns:
        dict: A dictionary containing the found action files categorized by their suffix.
    """
    root = action_root.absolute()
    name1 = name.replace(".", "/")
    dn = root / name1
    logging.debug(f"dn: {dn}")
    files = {}
    if dn.is_dir():
        files["dir"] = str(dn)
    glob_pattern = f"{name1}.*"
    logging.debug(f"calling root: {root}, pattern: {glob_pattern}")
    for fn in root.glob(glob_pattern):
        logging.debug(f"checking {fn}")
        if fn.is_file():
            suffix = "".join(fn.suffixes)[1:].replace(".", "_")
            if re.match("[a-z0-9_\.]+$", suffix, re.IGNORECASE):
                files.setdefault(suffix, str(fn))

    logging.debug(f"files found for action {name}: {files}")

    return files

def resolve_action_name(name, parent):
    """
    Resolve the full action name from a relative action name and its parent.

    Args:
        name (str): The action name to resolve.
        parent (str): The parent action name.

    Returns:
        str: The resolved full action name.

    Raises:
        ValueError: If the name is a relative name but no parent is provided.
    """
    if name.startswith("."):
        if parent:
            return f"{parent}{name}"
        else:
            raise ValueError("Can't resolve relative action name without a parent")
    return name

def lookup(name, parent=""):
    """
    Lookup and retrieve an action by its name.

    Args:
        name (str): The name of the action to look up.
        parent (str): An optional parent action name.

    Returns:
        Action: The action instance corresponding to the provided name.

    Raises:
        ValueError: If the action type is not declared or the action file is not found.
    """
    name = resolve_action_name(name, parent)
    if name not in _action_cache:
        actions_dir = pathlib.Path(cfg["fs.actions"])
        files = _find_action_files(actions_dir, name)

        cfg_path = "actions." + ".children.".join(name.split("."))
        acfg = cfg.cd(cfg_path)

        for ext in ("yaml", "json"):
            if ext in files:
                acfg.merge_file(files[ext], frame=-1)

        for k, v in files.items():
            acfg.setdefault(f"files.{k}", v)

        action_type = acfg.setdefault("type", _action_type_lookup(files))
        if action_type is None:
            raise ValueError(f"Action {name} has no type declared or action file not found")

        logging.debug(f"action_type: {action_type}")
        _action_cache[name] = _action_class_lookup(action_type)(name, acfg)

    return _action_cache[name]

def run(name):
    """
    Execute the action corresponding to the provided name.

    Args:
        name (str): The name of the action to run.
    """
    lookup(name).run()



from plpipes.plugin import Plugin, plugin
import plpipes.filesystem as fs
import pyspark.sql

def _clean_path(path):
    return str(path.resolve()).replace("\\", "/")

@plugin
class SparkEmbeddedPlugin(Plugin):

    def init_spark_session(self, ssc):
        home = fs.path(ssc.get("home", "spark"), mkdir=True)

        builder = pyspark.sql.SparkSession.builder \
                                          .appName(ssc.get("app_name", 'work')) \
                                          .config('spark.sql.warehouse.dir',
                                                  _clean_path(home / 'spark-warehouse')) \
                                          .config('javax.jdo.option.ConnectionURL',
                                                  f"jdbc:derby:;databaseName={_clean_path(home / 'metastore-db')};create=true") \
                                          .config('spark.logLevel', ssc.get('log_level', 'WARN')) \
                                          .config('spark.sql.catalogImplementation', 'hive')

        for k, v in ssc.to_flat_dict('extra').items():
            builder = builder.config(k, v)

        builder = builder.enableHiveSupport()

        return builder.getOrCreate()



"""
This module contains drivers for handling SQL-related actions within the plpipes framework.

The supported actions include executing raw SQL scripts, creating SQL tables from templates,
and creating SQL views from templates. Each action driver is responsible for loading the
appropriate SQL configuration, rendering templates, and executing the SQL commands
against the specified database.
"""

import logging
import pathlib

from plpipes.action.base import Action
from plpipes.action.registry import register_class

from plpipes.config import cfg

class _SqlTemplated(Action):
    """
    Base class for SQL templated actions.

    This class handles loading and rendering SQL templates, as well as managing
    the configuration and source file extraction.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the _SqlTemplated action.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.
        """
        super().__init__(*args, **kwargs)
        acfg, source = self._break_source_file()
        self._cfg.merge(acfg)
        self._source = source

    def _break_source_file(self):
        """
        Extracts the YAML header from the source file.

        Returns:
            A tuple containing the extracted configuration and the remaining
            lines of the source file.

        Raises:
            ValueError: If the YAML header is not properly closed.
        """
        # Extract the YAML header from the source file.
        # This is a hacky state machine.

        fn = self._source_fn()

        with open(fn, "r") as f:
            in_yaml = False
            yaml = []

            lines = f.readlines()
            ix = 0
            while ix < len(lines):
                sl = lines[ix].rstrip()
                ix += 1
                if sl == '---':
                    yaml_start = ix
                    while ix < len(lines):
                        sl = lines[ix].rstrip()
                        if sl == '---':
                            import yaml
                            acfg = yaml.safe_load("".join(lines[yaml_start:ix]))
                            lines = "".join(lines[:yaml_start] +
                                            ["--- YAML header removed.\n"] * (ix - yaml_start) +
                                            lines[ix:])
                            return acfg, lines
                        ix += 1
                    else:
                        raise ValueError("YAML header never closed")

                elif sl != '':
                    break
            return {}, "".join(lines)

    def do_it(self):
        """
        Executes the SQL code rendered from the source template.
        """
        sql_code = self._render_source_template()
        self._run_sql(sql_code)

    def _render_source_template(self):
        """
        Renders the SQL source template using the specified template engine.

        Returns:
            The rendered SQL code.

        Raises:
            ValueError: If an unsupported SQL template engine is specified.
        """
        engine = self._cfg.get("engine", "jinja2")

        if engine == "jinja2":
            from . import jinja2
            return jinja2.render_template(self._source, {'cfg': cfg, 'acfg': self._cfg, 'str': str})

        raise ValueError(f"Unsupported SQL template engine {engine}")

    def _short_name_to_table(self):
        """
        Converts the short name to a valid table name.

        Returns:
            A string representing the table name.
        """
        name = self.short_name()
        return name.replace("-", "_")

class _SqlTableCreator(_SqlTemplated):
    """
    Action for creating SQL tables from a template.
    """
    def _source_fn(self):
        """
        Retrieves the source file name for the table SQL.

        Returns:
            The source file path for the table SQL.
        """
        return self._cfg["files.table_sql"]

    def _run_sql(self, sql_code):
        """
        Executes the SQL code to create a table in the specified database.

        Args:
            sql_code: The SQL code to be executed.
        """
        import plpipes.database as db

        source_db = self._cfg.get("source_db", "work")
        target_db = self._cfg.get("target_db", "work")
        if source_db == target_db:
            db.create_table(self._short_name_to_table(), sql_code, db=source_db)
        else:
            iter = db.query(sql_code, db=source_db)
            db.create_table(self._short_name_to_table(), iter, db=target_db)

class _SqlViewCreator(_SqlTemplated):
    """
    Action for creating SQL views from a template.
    """
    def _source_fn(self):
        """
        Retrieves the source file name for the view SQL.

        Returns:
            The source file path for the view SQL.
        """
        return self._cfg["files.view_sql"]

    def _run_sql(self, sql_code):
        """
        Executes the SQL code to create a view in the database.

        Args:
            sql_code: The SQL code to be executed.
        """
        from plpipes.database import create_view
        create_view(self._short_name_to_table(), sql_code)

class _SqlRunner(_SqlTemplated):
    """
    Action for executing raw SQL scripts.
    """
    def _source_fn(self):
        """
        Retrieves the source file name for the SQL script.

        Returns:
            The source file path for the SQL script.
        """
        return self._cfg["files.sql"]

    def _run_sql(self, sql_code):
        """
        Executes the SQL code from a script.

        Args:
            sql_code: The SQL code to be executed.
        """
        from plpipes.database import execute_script
        execute_script(sql_code)

register_class("sql_script", _SqlRunner, "sql")
register_class("sql_table_creator", _SqlTableCreator, "table_sql", "table.sql")
register_class("sql_view_creator", _SqlViewCreator, "view_sql", "view.sql")



from pathlib import Path
import mkdocs_gen_files

# mkdocstrings is not able to parse these modules correctly!
skip = ["plpipes.database.driver.azure_sql",
        "plpipes.database.driver.plugin.azure_sql"]

# Ruta a tu código fuente (ajusta si no es "src")
src_root = Path("src")
# Nombre del paquete principal (ajusta a lo tuyo)

# Recorre todos los archivos .py dentro de src/
for former_path in sorted(src_root.rglob("*.py")):
    path = former_path.relative_to(src_root)
    if path.name == "__init__.py":
        path = path.parent # with_name("base")
    else:
        path = path.with_name(path.stem) # + "1") #with_suffix("")
    module = ".".join(path.parts)
    doc = path.with_suffix(".md")

    if module in skip:
        print(f"Skipping {module}")
        continue
    # print(f"Saving file {doc} for module {module}")

    with mkdocs_gen_files.open("reference/"+str(doc), "w") as f:
        f.write(f"""
# {module}
::: {module}
""")
    # Añade la entrada a SUMMARY.md (para literate-nav)
    with mkdocs_gen_files.open("reference/summary.md", "a") as nav:
        print(f"- [`{module}`]({doc})", file=nav)



import logging
import pathlib
import prql_python as prql

from plpipes.action.registry import register_class

from plpipes.action.driver.sql import _SqlTableCreator, _SqlRunner

class _PrqlTemplated:

    def _read_and_render_sql_template(self, fn):
        prql_code = super()._read_and_render_sql_template(fn)
        import prql
        return prql.to_sql(prql_code)

class _PrqlTableCreator(_PrqlTemplated, _SqlTableCreator):
    def _source_fn(self):
        return self._cfg["files.table_prql"]

class _PrqlViewCreator(_PrqlTemplated, _SqlViewCreator):
    def _source_fn(self):
        return self._cfg["files.view_prql"]

class _PrqlRunner(_PrqlTemplated, _SqlRunner):
    def _source_fn(self):
        return self._cfg["files.prql"]

register_class("prql_script", _PrqlRunner, "prql")
register_class("prql_table_creator", _PrqlTableCreator, "table_prql", "table.prql")
register_class("prql_view_creator", _PrqlTableCreator, "view_prql", "view.prql")



from plpipes.config import cfg
from pathlib import Path

"""
Module: plpipes.filesystem
This module provides functionalities for handling file paths, reading from and writing to various file formats,
and managing directories using the configuration settings defined in the application's config.
"""

def path(*args, mkparentdir=False, mkdir=False, **kwargs):
    """
    Generate a Path object based on the provided arguments.

    Args:
        *args: Positional arguments for the _path function.
        mkparentdir (bool): If True, create the parent directory of the path.
        mkdir (bool): If True, create the path directory.
        **kwargs: Additional keyword arguments for the _path function.

    Returns:
        Path: A Path object representing the constructed path.
    """
    p = _path(*args, **kwargs)
    if mkdir:
        p.mkdir(exist_ok=True, parents=True)
    elif mkparentdir:
        p.parent.mkdir(exist_ok=True, parents=True)
    return p

def _path(relpath=None, section=None):
    """
    Construct a Path object from a relative path and section.

    Args:
        relpath (str or None): Relative path to append to the base path.
        section (str or None): Configuration section to use.

    Returns:
        Path: A Path object representing the base path or a path with the relative path appended.
    """
    if section is None:
        section = "work"
    start = Path(cfg["fs." + section])
    if relpath is None:
        return start
    return start / relpath

def assign_section(target_section, relpath=None, section=None, **kwargs):
    """
    Assign a target section in the configuration and return a constructed path.

    Args:
        target_section (str): The section to assign.
        relpath (str or None): Relative path to use if section does not exist.
        section (str or None): Alternative section to use.
        **kwargs: Additional keyword arguments for the path function.

    Returns:
        Path: A Path object for the assigned section.
    """
    if target_section not in cfg.cd("fs"):
        if relpath is None:
            relpath = target_section
        cfg["fs." + target_section] = str(_path(relpath, section))
    return path(section=target_section, **kwargs)

def openfile(relpath, mode="r", section=None):
    """
    Open a file and return the file object.

    Args:
        relpath (str): Relative path to the file.
        mode (str): Mode in which to open the file (default is 'r').
        section (str or None): Configuration section to use.

    Returns:
        file object: The opened file object.
    """
    return open(path(relpath, section), mode)

def read_csv(relpath, section=None, **kwargs):
    """
    Read a CSV file and return a DataFrame.

    Args:
        relpath (str): Relative path to the CSV file.
        section (str or None): Configuration section to use.
        **kwargs: Additional keyword arguments for pandas read_csv.

    Returns:
        DataFrame: DataFrame containing the CSV data.
    """
    import pandas as pd
    return pd.read_csv(path(relpath, section), **kwargs)

def write_csv(relpath, df, section=None, mkdir=True, **kwargs):
    """
    Write a DataFrame to a CSV file.

    Args:
        relpath (str): Relative path for the CSV file.
        df (DataFrame): DataFrame to write.
        section (str or None): Configuration section to use.
        mkdir (bool): If True, create the directory if it does not exist.
        **kwargs: Additional keyword arguments for pandas to_csv.

    Returns:
        None
    """
    target = path(relpath, section)
    if mkdir:
        target.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(target, **kwargs)

def write_text(relpath, text, section=None, mkdir=True):
    """
    Write text to a file.

    Args:
        relpath (str): Relative path for the text file.
        text (str): Text to write to the file.
        section (str or None): Configuration section to use.
        mkdir (bool): If True, create the directory if it does not exist.

    Returns:
        int: Number of characters written to the file.
    """
    target = path(relpath, section)
    if mkdir:
        target.parent.mkdir(parents=True, exist_ok=True)
    with open(target, "w") as f:
        return f.write(text)

def read_text(relpath, section=None, encoding="utf-8"):
    """
    Read text from a file.

    Args:
        relpath (str): Relative path to the text file.
        section (str or None): Configuration section to use.
        encoding (str): Encoding to use for reading the file (default is 'utf-8').

    Returns:
        str: The contents of the file as a string.
    """
    target = path(relpath, section)
    with open(target, "r", encoding=encoding) as f:
        return f.read()

def write_yaml(relpath, data, section=None, mkdir=True, **kwargs):
    """
    Write data to a YAML file.

    Args:
        relpath (str): Relative path for the YAML file.
        data (any): Data to serialize to YAML.
        section (str or None): Configuration section to use.
        mkdir (bool): If True, create the directory if it does not exist.
        **kwargs: Additional keyword arguments for yaml.dump.

    Returns:
        None
    """
    import yaml
    target = path(relpath, section)
    if mkdir:
        target.parent.mkdir(parents=True, exist_ok=True)
    with open(target, "w") as f:
        yaml.dump(data, f, **kwargs)

def read_yaml(relpath, section=None):
    """
    Read data from a YAML file.

    Args:
        relpath (str): Relative path to the YAML file.
        section (str or None): Configuration section to use.

    Returns:
        any: The deserialized data from the YAML file.
    """
    import yaml
    with openfile(relpath, section=section) as f:
        return yaml.safe_load(f)

def read_json(relpath, section=None):
    """
    Read data from a JSON file.

    Args:
        relpath (str): Relative path to the JSON file.
        section (str or None): Configuration section to use.

    Returns:
        any: The deserialized data from the JSON file.
    """
    import json
    with openfile(relpath, section=section) as f:
        return json.load(f)

def tempdir(parent=None):
    """
    Create a temporary directory for file operations.

    Args:
        parent (str or None): Parent directory for the temporary directory.

    Returns:
        TemporaryDirectory: Temporary directory context manager.
    """
    if parent is None:
        parent = fs.path("tmp")
    import tempfile

    return tempfile.TemporaryDirectory(dir=parent)

def read_excel(relpath, section=None, **kwargs):
    """
    Read data from an Excel file and return a DataFrame.

    Args:
        relpath (str): Relative path to the Excel file.
        section (str or None): Configuration section to use.
        **kwargs: Additional keyword arguments for pandas read_excel.

    Returns:
        DataFrame: DataFrame containing the Excel data.
    """
    import pandas as pd
    return pd.read_excel(path(relpath, section), **kwargs)

def write_excel(relpath, df, section=None, mkparentdir=True, autofilter=False, **kwargs):
    """
    Write a DataFrame to an Excel file.

    Args:
        relpath (str): Relative path for the Excel file.
        df (DataFrame): DataFrame to write.
        section (str or None): Configuration section to use.
        mkparentdir (bool): If True, create the parent directory if it does not exist.
        autofilter (bool): If True, apply autofilter to the written Excel file.
        **kwargs: Additional keyword arguments for pandas to_excel.

    Returns:
        Path: The path of the written Excel file.
    """
    target = path(relpath, section=section, mkparentdir=mkparentdir)
    df.to_excel(target, index=False, **kwargs)

    if autofilter:
        import openpyxl
        wb = openpyxl.load_workbook(target)
        ws = wb.active
        ws.auto_filter.ref = ws.dimensions
        wb.save(target)

    return target



"""
plpipes.exceptions

This module defines custom exception classes for handling various
error scenarios in the plpipes package. These exceptions extend
the built-in Exception class and provide better context for error
handling and debugging.

"""

class AuthenticationError(Exception):
    """
    Exception raised for authentication-related errors.
    """
    pass

class CloudError(Exception):
    """
    Base exception class for all cloud-related errors.
    """
    pass

class CloudFSError(CloudError):
    """
    Exception raised for file system errors related to cloud operations.
    """
    pass

class CloudAccessError(CloudError):
    """
    Exception raised for access-related errors when interacting with cloud services.
    """
    pass



from plpipes.database.driver.filedb import FileDBDriver

from plpipes.database.driver.transaction import Transaction
from plpipes.database.sqlext import Wrap

import sqlalchemy as sa
import sqlalchemy.sql as sas

import plpipes.plugin
extension_register = plpipes.plugin.Registry("db_backend_sqlite_extension", "plpipes.database.driver.sqlite.extension")

class _SQLiteMapAsPandas:
    def __init__(self):
        self.rows = []

    def step(self, *row):
        self.rows.append(row)

    def finalize(self):
        try:
            df = pandas.DataFrame(self.rows, columns=list(args))
            r = self.process_pandas(df)
            return json.dumps(r)
        except Exception as ex:
            logging.error(f"Exception caught: {ex}")
            raise ex

class SQLiteTransaction(Transaction):
    def create_function(self, name, nargs, pyfunc):
        self._driver._create_function(self, name, nargs, pyfunc)

class SQLiteDriver(FileDBDriver):

    _transaction_factory = SQLiteTransaction

    def __init__(self, name, drv_cfg):
        super().__init__(name, drv_cfg, "sqlite")

        self._extensions = []
        for extension_name in drv_cfg.get("extensions", []):
            extension_class = extension_register.lookup(extension_name)
            self._extensions.append(extension_class(self, extension_name, drv_cfg))


    def _execute_script(self, txn, sql):
        import sqlparse
        for statement in sqlparse.split(sql):
            txn._conn.execute(Wrap(statement))

    def _create_function(self, txn, name, nargs, pyfunc):
        txn._conn.connection.create_function(name, nargs, pyfunc)

    def create_table_from_query_group_and_map(self,
                                              name, sql, by,
                                              function, args, out,
                                              if_exists):
        n_out = len(out)

        class _C(_SQLiteMapAsPandas):
            def process_pandas(self, df):
                r = function(df)
                if len(r) != n_out:
                    raise ValueError(f"Wrong number of items returned by function, {len(r)} found, {n_out} expected")
                return r

        key = self._next_key()
        agg_name = f"_map_to_pandas_aggregate_XXX{key}"
        out_name = f"_aggregate_output_column_XXX{key}"

        extractors = [f"""json_extract({out_name}, "$[{ix}]") as {col_name}"""
                      for ix, col_name in enumerate(out)]

        full_sql = f"""
        create table {name} as
          select {','.join([*by, *extractors])}
          from (
            select {','.join([*by, ''])}
                   {agg_name}({', '.join(args)}) as {out_name}
            from (
              {sql}
            )
            group by {', '.join(by)}
         )"""

        logging.debug(f"SQL code: {repr(full_sql)}")

        with self.connection() as conn:
            if if_exists:
                self.execute(f"drop table if exists {name}")
            conn.connection.create_aggregate(agg_name, len(args), _C)
            try:
                conn.execute(full_sql)
            finally:
                conn.connection.create_aggregate(agg_name, len(args), None)

    def query_group_and_map(self,
                            sql, by,
                            function, args):

        class _C(_MapAsPandas):
            def process_pandas(self, df):
                function(df)
                return 1

        key = self._next_key()
        agg_name = f"_map_to_pandas_aggregate_XXX{_last_key}"

        full_sql = f"""
select sum({agg_name}({', '.join(args)}))
from (
  {sql}
)
group by {', '.join(by)}
"""
        logging.debug(f"SQL code: {repr(full_sql)}")

        if exists:
            self.execute(f"drop table if exists {name}")

        with self.connection() as conn:
            conn.connection.create_aggregate(agg_name, len(args), _C)
            try:
                self.query(full_sql)
            finally:
                conn.connection.create_aggregate(agg_name, len(args), None)

    def _list_tables_query(self):
        return sas.select(sas.column("name")).select_from(sas.table("sqlite_master")).where(sas.column("type") == "table")

    def _list_views_query(self):
        return sas.select(sas.column("name")).select_from(sas.table("sqlite_master")).where(sas.column("type") == "view")


    def ibis(self):
        import ibis
        ic = ibis.sqlite.from_connection(self._engine.connect().connection)
        ic._plpipes_driver = self
        return ic



"""
This module provides a configuration management system. 
The main way to access the configuration is through the global variable `cfg`, 
which behaves like a dictionary but supports dotted keys for hierarchical 
access to configuration entries.
"""

import json
import logging
import re
import copy
import collections.abc

def _merge_any(tree, new):
    """Merge new configuration data into the existing tree structure."""
    if isinstance(new, dict):
        if isinstance(tree, dict):
            for k, v in new.items():
                tree[k] = _merge_any(tree[k], v) if k in tree else copy.deepcopy(v)
            return tree
        else:
            return {k: _merge_any(None, v) for k, v in new.items()}
    if isinstance(new, list):
        return [_merge_any(None, e) for e in new]
    if isinstance(new, (int, float, bool)) or new is None:
        return new
    return str(new)

def _flatten_tree(tree):
    """Flatten a nested dictionary into a single level dictionary with dotted keys."""
    flat = {}
    def rec(subtree, path):
        for k, v in subtree.items():
            child_path = path + [k]
            if isinstance(v, dict):
                rec(v, child_path)
            else:
                flat[".".join(child_path)] = v
    rec(tree, [])
    return flat


class ConfigStack:
    """Manage a stack of configuration frames."""

    def __init__(self):
        """Initialize a new ConfigStack with empty frames and cache."""
        self._frames = []
        self._cache = {}

    def _cd(self, path):
        """Change directory to the given path in the configuration stack."""
        return _Ptr(self, path)

    def root(self):
        """Return a pointer to the root of the configuration."""
        return self._cd("")

    def reset_cache(self):
        """Reset the caching mechanism for configuration retrieval."""
        self._cache = {}

    def _get(self, key, frame=0):
        """Get the value of a configuration key, with caching."""
        if frame == 0:
            if key not in self._cache:
                self._cache[key] = self._get_nocache(key, 0)
            return self._cache[key]
        else:
            return self._get_nocache(key, frame)

    def _get_nocache(self, key, frame):
        """
        Retrieve the value associated with a given configuration key
        while considering wildcard entries and frame specificity.

        The method implements a search algorithm that takes into account
        the specificity of configuration entries and their order of 
        loading based on frame. It follows these rules:

        1. More specific entries always win over less specific ones.
        2. For entries with the same specificity, the one from the
           lowest frame (last loaded) wins.

        This is implemented using a mix of A*/depth-first search algorithm.

        Parameters:
        key (str): The configuration key to look up, expressed in
                   dotted notation.
        frame: The frame context from which to retrieve the key.

        Returns:
        The value associated with the specified key.

        Raises:
        KeyError: If the key is not found in the configuration.
        ValueError: If the key does not point to a terminal node.
        """

        (key_part, *right) = key.split(".")
        # queue structure:
        #   specificity, frame_ix, tree, left_path, frozen_key, rigth_path
        queue = [(('!', '*')[k], ix, f, (key_part, '*')[k], (key_part, '*')[k], right)
                 for ix, f in enumerate(self._frames[frame:])
                 for k in (0, 1)]
        while queue:
            queue.sort(reverse=True)
            (specifity, frame_ix, tree, left, key_part, right) = queue.pop()
            while True:
                try:
                    if isinstance(tree, dict):
                        tree = tree[key_part]
                    elif isinstance(tree, list):
                        queue = []
                        if key_part.isnumeric():
                            tree = tree[int(key_part)]
                        else:
                            raise IndexError("Expecting numeric key")
                    else:
                        raise ValueError(f"Config key '{key}' traversing blocked by a non dictionary object at {left}")
                except (IndexError, KeyError):
                    break

                if right:
                    (key_part, *right) = right
                    queue.append((specifity + "*", frame_ix, tree, left + ".*", '*', right))
                    left = left + "." + key_part
                    specifity += "!"
                else:
                    if isinstance(tree, dict):
                        raise ValueError(f"config key '{key}' does not point to a terminal node")
                    return tree
        raise KeyError(f"config key '{key}' not found")

    def _contains(self, key):
        """Check if the configuration contains the specified key."""
        try:
            self._get(key, 0)
            return True
        except KeyError:
            return False

    def _merge(self, key, newtree, frame=0):
        """Merge new configuration data into the specified key of the frame."""
        # Auto-allocate frames
        if len(self._frames) <= frame:
            self._frames += [{} for _ in range(frame - len(self._frames) + 1)]
        tree = self._frames[frame]

        if key != "":
            parts = key.split(".")
            last = parts.pop()
            for p in parts:
                if (p not in tree) or (not isinstance(tree[p], dict)):
                    tree[p] = {}
                tree = tree[p]
            tree[last] = _merge_any(tree.get(last, None), newtree)
        else:
            if not isinstance(newtree, dict):
                raise ValueError("Top configuration must be a dictionary")
            self._frames[frame] = _merge_any(tree, newtree)
        self._cache = {}

    def _set(self, key, value):
        """Set the value for a given configuration key."""
        if not isinstance(value, (str, int, float, bool, list)) and value is not None:
            if isinstance(value, dict):
                raise ValueError("It is not possible to set a configuration entry to a dictionary, use merge instead")
            value = str(value)
        self._merge(key, value)

    def _multicd(self, key):
        """Change directory to a key, considering all matching frames."""
        # queue structure:
        #   specificity, frame_ix, tree
        queue = [("", ix, f) for ix, f in enumerate(self._frames)]
        if key != "":
            right = key.split(".")
            while right:
                queue.sort()
                key = right.pop(0)
                new_queue = []
                for specifity, ix, tree in queue:
                    if isinstance(tree, dict):
                        for s in ('!', '*'):
                            k = s if s == '*' else key
                            if k in tree:
                                new_queue.append((specifity + s, ix, tree[k]))
                    elif not new_queue:
                        raise ValueError(f"Config key '{key}' blocked by a non dictionary object")
                    else:
                        break
                queue = new_queue
        return [t for _, _, t in sorted(queue, reverse=True)]

    def _to_tree(self, key, defaults=None):
        """Convert a key's configuration to a tree structure."""
        m = self._multicd(key)
        tree = {}
        if defaults is not None:
            tree = _merge_any(tree, defaults)
        for other in m:
            tree = _merge_any(tree, other)
        return tree

    def _keys(self, key):
        """Retrieve the keys in a specified configuration key."""
        m = self._multicd(key)
        seen = set()
        inner_is_dict = True
        for other in m:
            if isinstance(other, dict):
                if not inner_is_dict:
                    inner_is_dict = True
                    seen = set()
                for k in other.keys():
                    if k != '*':
                        seen.add(k)
            else:
                inner_is_dict = False
        if inner_is_dict:
            return sorted(seen)
        raise ValueError(f"Config key '{key}' blocked by a non dictionary object")

    def _squash_frames(self):
        """Merge all frames into one."""
        tree = self._frames.pop()
        while self._frames:
            tree = _merge_any(tree, self._frames.pop())
        self._frames.append(tree)


class _Ptr(collections.abc.MutableMapping):
    """Pointer to a specific path in the configuration stack."""

    def __init__(self, stack, path):
        """Initialize a pointer to the given path in the specified stack."""
        self._path = path
        self._stack = stack

    def cd(self, key):
        """Change directory to the specified key in the pointer's path."""
        return self._stack._cd(self._mkkey(key)) if key else self

    def _mkkey(self, key):
        """Create a full key from the current path and the new key."""
        if key == "":
            return self._path
        if self._path == "":
            return key
        return f"{self._path}.{key}"

    def __getitem__(self, key):
        """Retrieve the value of the specified key."""
        return self._stack._get(self._mkkey(key))

    def __contains__(self, key):
        """Check if the pointer contains the specified key."""
        return self._stack._contains(self._mkkey(key))

    def __setitem__(self, key, value):
        """Set the value for the specified key in the pointer."""
        return self._stack._set(self._mkkey(key), value)

    def __delitem__(self, key):
        """Delete the specified key from the configuration."""
        self._stack._del(key)

    def __len__(self):
        """Return the number of keys in the pointer's path."""
        return len(self.__keys__())

    def to_tree(self, key="", defaults=None):
        """Convert the configuration in the pointer to a tree structure."""
        return self._stack._to_tree(self._mkkey(key), defaults)

    def to_flat_dict(self, key="", defaults=None):
        """Retrieve the configuration as a flattened dictionary."""
        t = self._stack._to_tree(self._mkkey(key), defaults)
        return _flatten_tree(t)

    def to_json(self, key="", defaults=None):
        """Convert the configuration to a JSON string."""
        return json.dumps(self.to_tree(key, defaults))

    def merge(self, tree, key="", frame=0):
        """Merge new configuration data into the specified key."""
        return self._stack._merge(self._mkkey(key), tree, frame)

    def merge_file(self, fn, key="", frame=0):
        """Merge configuration data from a file into the specified key."""
        logging.debug(f"Reading configuration file {fn}")
        with open(fn, "r+", encoding="utf8") as f:
            if re.search(r'\.ya?ml', str(fn), re.IGNORECASE):
                import yaml
                tree = yaml.safe_load(f)
            elif re.search(r'\.json', str(fn), re.IGNORECASE):
                import json
                tree = json.load(f)
            else:
                raise ValueError(f"Can't determine file type for {str(fn)}")
        self.merge(tree, key, frame=frame)

    def squash_frames(self):
        """Merge all frames in the stack."""
        self._stack._squash_frames()

    def __iter__(self):
        """Iterate over the keys in the current pointer's path."""
        for k in self._stack._keys(self._mkkey("")):
            yield k

    def __keys__(self):
        """Retrieve the keys of the current pointer's path."""
        return self._stack._keys(self._mkkey(""))

    def __str__(self):
        """Return a string representation of the configuration at the pointer's path."""
        return str(self._stack._to_tree(""))

    def copydefaults(self, src, *keys, **keys_with_default):
        """Copy default configuration values from the source."""
        for key in keys:
            if key not in self:
                if key in src:
                    self[key] = src[key]
                else:
                    logging.debug(f"key {key} not found!")
        for key, default in keys_with_default.items():
            if key not in self:
                self[key] = src.get(key, default)

    def setdefault_lazy(self, key, cb):
        """Set a default value for a key lazily using a callback."""
        if key not in self:
            self[key] = cb()
        return self[key]

    def getany(self, *keys, default=None):
        """Retrieve the first available value from the specified keys."""
        for key in keys:
            try:
                return self[key]
            except KeyError:
                pass
        return default

cfg_stack = ConfigStack()
cfg = cfg_stack.root()
"""Singleton instance of the configuration object provided by PLPipes.
This object allows access to the configuration settings managed by
plpipes and this module (plpipes.config). It provides a dictionary-like
interface to retrieve and manipulate configuration values."""
